import argparse
import asyncio
from ..client_auth import add_auth_to_argparse, create_rest_client


async def main():
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(title='Cred Types', description='Register a credential', required=True)

    def generic_args(p):
        p.add_argument('url', default='https://data.icecube.aq', help='data base url')
        g = p.add_mutually_exclusive_group(required=True)
        g.add_argument('--user', help='username to assign cred to')
        g.add_argument('--group', help='group to assign cred to')

    parser_s3 = subparsers.add_parser('S3', aliases=['s3'], help='Register an S3 credential')
    parser_s3.set_defaults(type_='s3')
    generic_args(parser_s3)
    parser_s3.add_argument('access_key', metavar='access-key', help='s3 access key')
    parser_s3.add_argument('secret_key', metavar='secret-key', help='s3 secret key')
    parser_s3.add_argument('bucket', action='append', help='bucket(s) available (use multiple times for more buckets)')

    parser_oauth = subparsers.add_parser('OAuth', aliases=['oauth'], help='Register an OAuth credential', description='Must include either an access or refresh token')
    parser_oauth.set_defaults(type_='oauth')
    generic_args(parser_oauth)
    parser_oauth.add_argument('--access-token', dest='access_token', help='access token')
    parser_oauth.add_argument('--refresh-token', dest='refresh_token', help='refresh token')
    parser_oauth.add_argument('--expire-date', dest='expire_date', type=float, default=None, help='(optional) manual expiration date in unix time')
    parser_oauth.add_argument('--last-use', dest='last_use', type=float, default=None, help='(optional) manual last use date in unix time')

    add_auth_to_argparse(parser)
    args = parser.parse_args()

    if args.type_ == 'oauth' and not (args.access_token or args.refresh_token):
        raise argparse.ArgumentError(argument=args.refresh_token, message='--access-token or --refresh-token is required')

    args.rest_url = 'https://credentials.iceprod.icecube.aq'
    rc = create_rest_client(args)

    if args.user:
        url = f'/users/{args.user}/credentials'
    else:
        url = f'/groups/{args.group}/credentials'

    data = {
        'url': args.url,
        'type': args.type_,
    }

    if args.type_ == 's3':
        data.update({
            'access_key': args.access_key,
            'secret_key': args.secret_key,
            'buckets': args.bucket,
        })
    elif args.type_ == 'oauth':
        if args.access_token:
            data['access_token'] = args.access_token
        if args.refresh_token:
            data['refresh_token'] = args.refresh_token
        if args.expire_date:
            data['expire_date'] = args.expire_date
        if args.last_use:
            data['last_use'] = args.last_use
    else:
        raise RuntimeError('bad type')

    await rc.request('POST', url, data)


if __name__ == '__main__':
    asyncio.run(main())



"""
Task submission directly to condor.
"""
import os
import logging
import getpass
from datetime import datetime,timedelta
import subprocess
import asyncio
from functools import partial
import gzip

import requests.exceptions

import iceprod
from iceprod.core import constants
from iceprod.core.exe import Config
from iceprod.core.resources import sanitized_requirements, rounded_requirements
from iceprod.core.exe_json import ServerComms
from iceprod.server import grid
from iceprod.server.globus import SiteGlobusProxy

logger = logging.getLogger('plugin-condor_direct')


async def check_call(*args, **kwargs):
    logger.info('subprocess_check_call: %r', args)
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    if await p.wait():
        raise Exception(f'command failed, return code {p.returncode}')
    return p


async def check_call_clean_env(*args, **kwargs):
    logger.info('subprocess_check_call: %r', args)
    env = os.environ.copy()
    del env['LD_LIBRARY_PATH']
    kwargs['env'] = env
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    if await p.wait():
        raise Exception(f'command failed, return code {p.returncode}')
    return p


async def check_output(*args, **kwargs):
    kwargs['stdout'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.STDOUT
    logger.info('subprocess_check_output: %r', args)
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    out,_ = await p.communicate()
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return out.decode('utf-8')


async def check_output_clean_env(*args, **kwargs):
    logger.info('subprocess_check_output: %r', args)
    kwargs['stdout'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.PIPE
    env = os.environ.copy()
    del env['LD_LIBRARY_PATH']
    kwargs['env'] = env
    p = await asyncio.create_subprocess_exec(*args, **kwargs)
    out,err = await p.communicate()
    out = out+b'\n'+err
    if p.returncode:
        raise Exception(f'command failed, return code {p.returncode}')
    return out.decode('utf-8')


class MyServerComms(ServerComms):
    def __init__(self, rest_client):
        self.rest = rest_client


class TaskInfo(dict):
    def __init__(self, **kwargs):
        self['dataset_id'] = None
        self['job_id'] = None
        self['task_id'] = None
        self['config'] = None
        self['pilot'] = None
        self['submit_dir'] = None
        super(TaskInfo, self).__init__(**kwargs)


def condor_os_reqs(os_arch):
    """Convert from OS_ARCH to Condor OS requirements"""
    os_arch = os_arch.rsplit('_',2)[0].rsplit('.',1)[0]
    reqs = 'OpSysAndVer =?= "{}"'.format(os_arch.replace('RHEL','CentOS').replace('_',''))
    reqs = reqs + '|| OpSysAndVer =?= "{}"'.format(os_arch.replace('RHEL','SL').replace('_',''))
    reqs = reqs + ' || OSGVO_OS_STRING =?= "{}"'.format(os_arch.replace('_',' '))
    return '('+reqs+')'


def run_async(func,*args,**kwargs):
    async def myfunc():
        return_obj = {'args':args,'kwargs':kwargs}
        try:
            return_obj['return'] = await func(*args,**kwargs)
        except Exception as e:
            logger.info('error running async', exc_info=True)
            return_obj['exception'] = e
        return return_obj
    return asyncio.ensure_future(myfunc())


def read_filename(filename):
    """Read a file that may potentially be gzipped"""
    data = ''
    if os.path.exists(filename):
        with open(filename) as f:
            data = f.read()
    elif os.path.exists(filename+'.gz'):
        try:
            with gzip.open(filename+'.gz', 'rt', encoding='utf-8') as f:

                try:
                    while True:
                        pos = f.tell()
                        ret = f.read(256)
                        if not ret:
                            break
                        data += ret
                except EOFError:
                    f.seek(pos)
                    try:
                        while True:
                            ret = f.read(1)
                            if not ret:
                                break
                            data += ret
                    except EOFError:
                        pass
        except Exception:
            logger.info('error reading gzip file', exc_info=True)
    if len(data) > 10**8:
        logger.warning(f'logfile {filename} has length {len(data)} and will be trimmed')
        data = data[-1*10**8:]
    return data


class condor_direct(grid.BaseGrid):
    """Plugin Overrides for HTCondor direct job submission"""

    batch_site = 'CondorDirect'
    batch_outfile = 'condor.out'
    batch_resources = {}

    def __init__(self, *args, **kwargs):
        super(condor_direct, self).__init__(*args, **kwargs)
        self.x509proxy = SiteGlobusProxy()

        # queue requirements
        self.site = self.batch_site
        if 'site' in self.queue_cfg:
            self.site = self.queue_cfg['site']
        self.resources = self.batch_resources.copy()
        self.resources['site'] = self.site
        if 'gpu' in self.site.lower():
            self.resources['gpu'] = 1
        elif 'cpu' in self.site.lower():
            self.resources['gpu'] = 0
        self.queue_params = {}
        if 'exclusive' in self.queue_cfg and self.queue_cfg['exclusive']:
            self.queue_params['requirements.site'] = self.site
        logger.info('resources: %r', self.resources)
        logger.info('queue params: %r', self.queue_params)

        self.grid_remove_once = set()

    async def upload_logfiles(self, task_id, dataset_id, submit_dir=None, reason=''):
        """
        Upload logfiles

        Args:
            task_id (str): task id
            dataset_id (str): dataset id
            submit_dir (str): (optional) submit dir for task
            reason (str): (optional) reason to inject into stdlog if it does not exist

        Returns:
            payload_failure (bool): indicate if the task had a payload failure
        """
        if submit_dir is None:
            submit_dir = ''

        payload_failure = False

        data = {'name': 'stdlog', 'task_id': task_id, 'dataset_id': dataset_id}

        # upload stdlog
        data['data'] = read_filename(os.path.join(submit_dir, constants['stdlog']))
        for line in data['data'].split('\n'):
            if 'task exe' in line and 'return code' in line:
                return_code = int(line.rsplit(':', 1)[1].strip())
                if return_code != 0 and return_code != 132:  # ignore SIGILL
                    payload_failure = True
                    break

        if not data['data']:
            data['data'] = reason
        await self.rest_client.request('POST', '/logs', data)

        # upload stderr
        data['name'] = 'stderr'
        data['data'] = read_filename(os.path.join(submit_dir, constants['stderr']))
        await self.rest_client.request('POST', '/logs', data)
        if payload_failure:
            for line in data['data'].split('\n'):
                # find cases where it's probably a node failure
                if ('No such file or directory' in line
                        or 'No space left on device' in line
                        or 'Illegal instruction' in line
                        or 'Input/output error' in line
                        or ('Killed' in line and 'env-shell.sh' in line)
                        or ('python: command not found' in line and 'env-shell.sh' in line)
                        or 'py3-v4.1.1/RHEL_8_x86_64/lib/libCore.so.6.18: undefined symbol: usedToIdentifyRootClingByDlSym' in line
                        or 'OpenCL ERROR: clGetPlatformIDs' in line):
                    payload_failure = False
                    break

        # upload stdout
        data['name'] = 'stdout'
        data['data'] = read_filename(os.path.join(submit_dir, constants['stdout']))
        await self.rest_client.request('POST', '/logs', data)

        return payload_failure

    async def get_hold_reason(self, submit_dir, resources=None):
        """Search for a hold reason in the condor.log"""
        if submit_dir is None:
            submit_dir = ''
        reason = None
        submit_filename = os.path.join(submit_dir, 'condor.submit')
        submit_data = {}
        if os.path.exists(submit_filename):
            with open(submit_filename) as f:
                for line in f:
                    line = line.strip().lower()
                    if '=' in line:
                        key, value = line.split('=', 1)
                        submit_data[key.strip()] = value.strip()
        filename = os.path.join(submit_dir, 'condor.log')
        if os.path.exists(filename):
            with open(filename) as f:
                for line in f:
                    line = line.strip().lower()
                    if 'policy violation' in line:
                        resource_type = None
                        val = 0
                        if 'memory limit' in line:
                            resource_type = 'memory'
                            try:
                                val = float(line.split('used')[-1].split('mb')[0].strip())/1024.
                            except Exception:
                                try:
                                    val = float(line.split(':')[-1].split('mb')[0].strip())/1024.
                                except Exception:
                                    pass
                        elif 'memory usage exceeded' in line:
                            resource_type = 'memory'
                        elif 'cpu limit' in line or 'cpu consumption limit':
                            resource_type = 'cpu'
                            try:
                                val = float(line.split('used')[-1].split('cores')[0].strip())
                            except Exception:
                                try:
                                    val = float(line.split('used')[-1].split('usr')[0].strip())
                                except Exception:
                                    pass
                        elif 'cpu usage exceeded' in line:
                            resource_type = 'cpu'
                        elif 'execution time limit' in line:
                            resource_type = 'time'
                            try:
                                val = float(line.split('used')[-1].split('.')[0].strip())/3600.
                            except Exception:
                                pass
                        elif 'local storage limit' in line:
                            resource_type = 'disk'
                            try:
                                val = float(line.split('used')[-1].split('mb')[0].strip())/1024.
                            except Exception:
                                try:
                                    val = float(line.split('used')[-1].split('gb')[0].strip())
                                except Exception:
                                    pass
                        elif 'disk usage exceeded' in line:
                            resource_type = 'disk'
                        if resource_type:
                            if val:
                                resources[resource_type] = val
                            reason = f'Resource overusage for {resource_type}: '
                            if resource_type in resources:
                                reason += f'{resources[resource_type]}'
                            break
                    elif 'cpu usage exceeded request_cpus' in line:
                        reason = 'Resource overusage for cpu: '
                        try:
                            reason += str(int(submit_data['request_cpus']))
                        except Exception:
                            pass
                        break
                    elif 'memory usage exceeded request_memory' in line:
                        reason = 'Resource overusage for memory: '
                        try:
                            reason += str(float(submit_data['request_memory'])/1000.)
                        except Exception:
                            pass
                        break
                    elif 'disk usage exceeded request_disk' in line:
                        reason = 'Resource overusage for disk: '
                        try:
                            reason += str(float(submit_data['request_disk'])/1000000.)
                        except Exception:
                            pass
                        break
                    elif 'runtime exceeded maximum' in line:
                        reason = 'Resource overusage for time: '
                        try:
                            reason += str(float(line.split()[-2].strip('('))/3600)
                        except Exception:
                            pass
                        break
                    elif 'Transfer output files failure' in line:
                        reason = 'Failed to transfer output files'
                        break
                    elif 'Transfer input files failure' in line:
                        reason = 'Failed to transfer input files'
                        break
                    elif 'failed due to remote transfer hook error' in line:
                        if 'failed to send file' in line:
                            reason = 'Failed to transfer output files'
                        elif 'failed to receive file' in line:
                            reason = 'Failed to transfer input files'
                        else:
                            reason = 'Failed to transfer files'
                        break
        return reason

    async def task_error(self, task_id, dataset_id, submit_dir, reason='',
                         site=None, pilot_id=None, kill=False, failed=False):
        """reset a task"""
        if submit_dir is None:
            submit_dir = ''
        # search for resources in stdout
        resources = {}
        batch_job_id = None
        filename = os.path.join(submit_dir, self.batch_outfile)
        if os.path.exists(filename):
            with open(filename) as f:
                resource_lines = False
                for line in f:
                    line = line.strip()
                    if (not batch_job_id) and 'Job submitted from host' in line:
                        batch_job_id = '.'.join(line.split('(', 1)[1].split('.')[0:2])
                    if line == 'Resources:':
                        resource_lines = True
                        continue
                    if resource_lines:
                        if ':' in line:
                            name,value = line.split(':',1)
                            if value.isdigit():
                                resources[name] = int(value)
                            else:
                                try:
                                    resources[name] = float(value)
                                except Exception:
                                    resources[name] = value
                            continue
                        else:
                            break

        if batch_job_id:
            resources.update(await self.get_grid_resources(batch_job_id))

        if (not reason) and failed:
            reason = 'payload failure'
        if not reason:
            # search for reason in logfile
            filename = os.path.join(submit_dir, constants['stdlog'])
            for line in read_filename(filename):
                line = line.strip()
                if 'failed to download' in line:
                    reason = 'failed to download input file(s)'
                if 'failed to upload' in line:
                    reason = 'failed to upload output file(s)'
                if 'Exception' in line and not reason:
                    reason = line
                if 'return code:' in line and not reason:
                    reason = 'task error: return code '+line.rsplit(':',1)[-1]
        if not reason:
            # check the batch system logs
            reason = await self.get_hold_reason(submit_dir, resources=resources)
        if not reason:
            reason = 'unknown failure'

        if not site:
            site = self.site

        comms = MyServerComms(self.rest_client)
        if kill:
            host = None
            filename = os.path.join(submit_dir, 'condor.log')
            if os.path.exists(filename):
                with open(filename) as f:
                    for line in f:
                        line = line.strip()
                        if 'Job executing on host' in line:
                            host = line.split('<')[-1].split(':')[0]
                        if 'Error from' in line:
                            host = line.split()[2].strip(':')
            submitter = grid.get_host()
            message = reason + f'\n\npilot_id: {pilot_id}\nhostname: {host}\nsubmitter: {submitter}\nsite: {site}'
            await comms.task_kill(task_id, dataset_id=dataset_id, reason=reason,
                                  resources=resources, message=message, site=site)
        else:
            await comms.task_error(task_id, dataset_id=dataset_id, reason=reason,
                                   resources=resources, site=site, failed=failed)

    async def finish_task(self, task_id, dataset_id, submit_dir, site=None):
        """complete a task"""
        if submit_dir is None:
            submit_dir = ''
        # search for reasources in slurm stdout
        resources = {}
        filename = os.path.join(submit_dir, self.batch_outfile)
        if os.path.exists(filename):
            with open(filename) as f:
                resource_lines = False
                for line in f:
                    line = line.strip()
                    if line == 'Resources:':
                        resource_lines = True
                        continue
                    if resource_lines:
                        if ':' in line:
                            name,value = line.split(':',1)
                            if value.isdigit():
                                resources[name] = int(value)
                            else:
                                try:
                                    resources[name] = float(value)
                                except Exception:
                                    resources[name] = value
                            continue
                        else:
                            break

        if not site:
            site = self.site

        comms = MyServerComms(self.rest_client)
        await comms.finish_task(task_id, dataset_id=dataset_id,
                                resources=resources, site=site)

    async def check_and_clean(self):
        """Check and clean the grid"""
        host = grid.get_host()

        # get time limits
        try:
            queued_time = timedelta(seconds=self.queue_cfg['max_task_queued_time'])
        except Exception:
            queued_time = timedelta(seconds=86400*2)
        try:
            processing_time = timedelta(seconds=self.queue_cfg['max_task_processing_time'])
        except Exception:
            processing_time = timedelta(seconds=86400*2)
        try:
            suspend_time = timedelta(seconds=self.queue_cfg['suspend_submit_dir_time'])
        except Exception:
            suspend_time = timedelta(seconds=86400)
        all_time = queued_time + processing_time + suspend_time
        time_dict = {'queued': queued_time,
                     'processing': queued_time+processing_time,
                     'completed': all_time,
                     'error': all_time,
                     'unknown': all_time}
        for t in time_dict:
            logger.debug("time limit: %s - %s",t,time_dict[t])
        now = datetime.utcnow()

        # get pilots from iceprod
        args = {
            'queue_host': host,
            'host': self.site,
            'keys': 'pilot_id|queue_host|grid_queue_id|submit_date|tasks',
        }
        ret = await self.rest_client.request('GET', '/pilots', args)

        # filter by queue host
        # index by grid_queue_id
        pilots = {}
        for pilot_id in ret:
            if (ret[pilot_id]['queue_host'] == host
                    and 'grid_queue_id' in ret[pilot_id]
                    and ret[pilot_id]['grid_queue_id']):
                pilots[ret[pilot_id]['grid_queue_id']] = ret[pilot_id]

        # first, check if anything has completed successfully
        # important to get grid status before completions, so we don't reset
        # newly completed jobs
        grid_jobs = await self.get_grid_status()
        grid_history = await self.get_grid_completions()

        logger.debug("iceprod pilots: %r", list(pilots))
        logger.debug("grid jobs: %r", list(grid_jobs))
        logger.debug("grid history: %r", list(grid_history))

        async def post_process(task):
            task_id = task['task_id']
            ret = await self.rest_client.request('GET', f'/tasks/{task_id}')
            if ret['status'] == 'processing':
                task['dataset_id'] = ret['dataset_id']

                logger.info('uploading logs for task %s', task_id)
                payload_failure = await self.upload_logfiles(task_id, task['dataset_id'],
                                                             submit_dir=task['submit_dir'])
                if task['grid']['status'] == 'ok':
                    # upload files (may be a no-op)
                    await self.upload_output(task)

                    logger.info('finished task %s', task_id)
                    await self.finish_task(task_id,
                                           dataset_id=task['dataset_id'],
                                           submit_dir=task['submit_dir'],
                                           site=task['grid']['site'])
                else:
                    logger.info('error in task %s', task_id)
                    if payload_failure:
                        logger.info('payload failed')
                    await self.task_error(task_id, task['dataset_id'],
                                          submit_dir=task['submit_dir'],
                                          site=task['grid']['site'],
                                          failed=payload_failure)

        async def post_process_complete(fut):
            ret = await fut
            task = ret['args'][0]
            if 'exception' in ret:
                reason = f'failed post-processing task\n{ret["exception"]}'
                logger.warning(reason)
                await self.upload_logfiles(task['task_id'],
                                           dataset_id=task['dataset_id'],
                                           submit_dir=task['submit_dir'],
                                           reason=reason)
                await self.task_error(task['task_id'],
                                      dataset_id=task['dataset_id'],
                                      submit_dir=task['submit_dir'],
                                      reason=reason)

        if grid_history:
            pilots_to_delete = {}
            awaitables = set()
            for gid in pilots:
                if gid in grid_history:
                    pilot_id = None
                    try:
                        pilot = pilots[gid]
                        pilot_id = pilot['pilot_id']
                        pilot['submit_dir'] = grid_history[gid]['submit_dir']
                        if pilot['tasks']:
                            task_id = pilot['tasks'][0]
                            logger.info('post-processing task %s', task_id)
                            task = TaskInfo(task_id=task_id, pilot=pilot,
                                            submit_dir=pilot['submit_dir'],
                                            grid=grid_history[gid])
                            awaitables.add(run_async(post_process, task))
                            while len(awaitables) >= 10:
                                done, pending = await asyncio.wait(awaitables, return_when=asyncio.FIRST_COMPLETED)
                                awaitables = pending
                                for fut in done:
                                    await post_process_complete(fut)
                    except Exception:
                        logger.error('error handling task', exc_info=True)

                    if pilot_id:
                        logger.info('deleting completed pilot %s, with gid %s', pilot_id, gid)
                        pilots_to_delete[pilot_id] = gid

            for fut in asyncio.as_completed(awaitables):
                await post_process_complete(fut)

            for pilot_id in pilots_to_delete:
                try:
                    await self.rest_client.request('DELETE', f'/pilots/{pilot_id}')
                except Exception:
                    logger.info('delete pilot error', exc_info=True)
                gid = pilots_to_delete[pilot_id]
                if gid in pilots:
                    del pilots[gid]

        # Now do the regular check and clean
        reset_pilots = set(pilots).difference(grid_jobs)
        prechecked_dirs = set()

        # give two attempts to find a grid job before removing it
        remove_grid_jobs = set()
        remove_once = set()
        for gid in set(grid_jobs).difference(pilots):
            if gid in self.grid_remove_once:
                remove_grid_jobs.add(gid)
            else:
                remove_once.add(gid)
        self.grid_remove_once = remove_once

        # check the queue
        grid_idle = 0
        for grid_queue_id in set(grid_jobs).intersection(pilots):
            status = grid_jobs[grid_queue_id]['status']
            submit_time = pilots[grid_queue_id]['submit_date']
            if '.' in submit_time:
                submit_time = datetime.strptime(submit_time, '%Y-%m-%dT%H:%M:%S.%f')
            else:
                submit_time = datetime.strptime(submit_time, '%Y-%m-%dT%H:%M:%S')

            if now - submit_time > time_dict[status]:
                logger.info('pilot over time: %r', pilots[grid_queue_id]['pilot_id'])
                # reset_pilots.add(grid_queue_id)
                remove_grid_jobs.add(grid_queue_id)
            elif status == 'error':
                logger.info('job error. pilot_id: %r, grid_id: %r',
                            pilots[grid_queue_id]['pilot_id'], grid_queue_id)
                # reset_pilots.add(grid_queue_id)
                remove_grid_jobs.add(grid_queue_id)

                pilot = pilots[grid_queue_id]
                if pilot['tasks']:
                    task_id = pilot['tasks'][0]
                    logger.info('post-processing task %s', task_id)
                    ret = await self.rest_client.request('GET', f'/tasks/{task_id}')
                    if ret['status'] == 'processing':
                        await self.task_error(task_id, ret['dataset_id'], pilot_id=pilot['pilot_id'],
                                              submit_dir=grid_jobs[grid_queue_id]['submit_dir'],
                                              kill=True, site=grid_jobs[grid_queue_id]['site'])
            elif status == 'queued':
                grid_idle += 1

            submit_dir = grid_jobs[grid_queue_id]['submit_dir']
            if submit_dir:
                # queueing systems don't like deleteing directories they know
                # about, so put them on a list of "don't touch"
                prechecked_dirs.add(submit_dir)
        self.grid_idle = grid_idle
        self.grid_processing = len(pilots)-len(reset_pilots)-grid_idle

        # check submit directories
        delete_dirs = set()
        for x in os.listdir(self.submit_dir):
            d = os.path.join(self.submit_dir,x)
            if d in prechecked_dirs:
                continue
            if os.path.isdir(d) and '_' in x:
                logger.debug('found submit_dir %s',d)
                mtime = datetime.utcfromtimestamp(os.path.getmtime(d))
                # use all_time instead of suspend_time because the
                # dir will have the submit time, not the last time
                if now-mtime < all_time:
                    continue  # skip for suspended or failed tasks
                delete_dirs.add(d)

        logger.info('%d processing pilots', self.grid_processing)
        logger.info('%d queued pilots', self.grid_idle)
        logger.info('%d ->reset', len(reset_pilots))
        logger.info('%d ->grid remove', len(remove_grid_jobs))
        logger.info('%d ->submit clean', len(delete_dirs))
        self.statsd.gauge('processing_pilots', self.grid_processing)
        self.statsd.gauge('queued_pilots', self.grid_idle)
        self.statsd.incr('reset_pilots', len(reset_pilots))
        self.statsd.incr('grid_remove', len(remove_grid_jobs))
        self.statsd.incr('clean_dirs', len(delete_dirs))

        if reset_pilots:
            logger.info('reset %r',reset_pilots)
            for grid_queue_id in reset_pilots:
                try:
                    pilot_id = pilots[grid_queue_id]['pilot_id']
                    await self.rest_client.request('DELETE', '/pilots/{}'.format(pilot_id))
                except KeyError:
                    pass
                except requests.exceptions.HTTPError as e:
                    if e.response.status_code == 404:
                        continue  # a missing pilot is the point of deleting it
                    logger.info('delete pilot error', exc_info=True)
                except Exception:
                    logger.info('delete pilot error', exc_info=True)

        # remove grid tasks
        if remove_grid_jobs:
            logger.info('remove %r',remove_grid_jobs)
            await asyncio.ensure_future(self.remove(remove_grid_jobs))

        if delete_dirs:
            await asyncio.ensure_future(self._delete_dirs(delete_dirs))

    async def queue(self):
        """Submit a pilot for each task, up to the limit"""
        host = grid.get_host()

        dataset_cache = {}

        async def make_pilot(task):
            """take a TaskInfo as input"""
            resources = self.resources.copy()

            # get full job, dataset, config info
            job = await self.rest_client.request('GET', f'/jobs/{task["job_id"]}')
            if task['dataset_id'] in dataset_cache:
                dataset, config = dataset_cache[task['dataset_id']]
            else:
                dataset = await self.rest_client.request('GET', f'/datasets/{task["dataset_id"]}')
                config = await self.rest_client.request('GET', f'/config/{task["dataset_id"]}')
                dataset_cache[task['dataset_id']] = (dataset, config)

            # update config with options, and parse it
            config['dataset'] = dataset['dataset']
            task.update({
                'config': config,
                'job': job['job_index'],
                'dataset': dataset['dataset'],
                'jobs_submitted': dataset['jobs_submitted'],
                'tasks_submitted': dataset['tasks_submitted'],
                'debug': dataset['debug'],
            })
            config = self.create_config(task)
            parser = Config(config)
            config = parser.parseObject(config, {})
            task['config'] = config

            task_cfg = None
            for t in config['tasks']:
                if t['name'] == task['name']:
                    task_cfg = t
                    break
            else:
                raise Exception(f'cannot find task in config for {task["task_id"]}')

            # get task requirements
            if 'requirements' in task:
                reqs = sanitized_requirements(task['requirements'])
            else:
                reqs = self.resources.copy()
            if task_cfg and 'requirements' in task_cfg:
                config_reqs = sanitized_requirements(task_cfg['requirements'])
                for k in config_reqs:
                    if k in reqs:
                        if isinstance(config_reqs[k], (int, float)) and reqs[k] < config_reqs[k]:
                            reqs[k] = config_reqs[k]
                        elif isinstance(config_reqs[k], str):
                            reqs[k] = config_reqs[k]
                    else:
                        reqs[k] = config_reqs[k]
            reqs = rounded_requirements(reqs)
            task['requirements'] = reqs

            # get task files
            if 'task_files' in task_cfg and task_cfg['task_files']:
                logger.info("getting task_files for %s", task['task_id'])
                comms = MyServerComms(self.rest_client)
                files = await comms.task_files(task['dataset_id'],
                                               task['task_id'])
                task_cfg['data'].extend(files)
                task_cfg['task_files'] = False

            # task config customizations
            await self.customize_task_config(task_cfg, job_cfg=task['config'], dataset=dataset)

            # create pilot
            if 'time' in resources:
                resources_available = {'time': resources['time']}
            else:
                resources_available = {'time': 1}
            resources_claimed = {}
            for k in ('cpu','gpu','memory','disk'):
                if k in resources and k in reqs:
                    resources_available[k] = resources[k]-reqs[k]
                else:
                    resources_available[k] = 0
                if k in reqs:
                    resources_claimed[k] = reqs[k]
                elif k in resources:
                    resources_claimed[k] = resources[k]
                else:
                    resources_claimed[k] = 0
            pilot = {
                'resources': resources,
                'resources_available': resources_available,
                'resources_claimed': resources_claimed,
                'tasks': [task['task_id']],
                'queue_host': host,
                'queue_version': iceprod.__version__,
                'host': self.site,
                'version': iceprod.__version__,
            }
            ret = await self.rest_client.request('POST', '/pilots', pilot)
            pilot['pilot_id'] = ret['result']
            task['pilot'] = pilot

            await self.setup_submit_directory(task)
            task['pilot']['submit_dir'] = task['submit_dir']

            # download files (may be a no-op)
            await self.download_input(task)

            # submit to queue
            await self.submit(task['pilot'])

            # update pilot
            pilot_id = task['pilot']['pilot_id']
            args = {'grid_queue_id': task['pilot']['grid_queue_id']}
            await self.rest_client.request('PATCH', f'/pilots/{pilot_id}', args)

        awaitables = set()
        queue_num = self.get_queue_num()
        args = {
            'requirements': self.resources.copy(),
            'query_params': self.queue_params,
        }
        logger.info(f'attempting to queue {queue_num} tasks, with args {args}')
        for _ in range(queue_num):
            # get a processing task
            try:
                ret = await self.rest_client.request('POST', '/task_actions/process', args)
            except Exception:
                logger.info('no more tasks to queue')
                break
            awaitables.add(run_async(make_pilot, TaskInfo(**ret)))

        grid_queue_ids = []
        for fut in asyncio.as_completed(awaitables):
            ret = await fut
            task = ret['args'][0]

            if 'exception' in ret:
                reason = f'failed queue task\n{ret["exception"]}'
                logger.warning(reason)
                await self.upload_logfiles(task['task_id'],
                                           dataset_id=task['dataset_id'],
                                           submit_dir=task['submit_dir'],
                                           reason=reason)
                await self.task_error(task['task_id'],
                                      dataset_id=task['dataset_id'],
                                      submit_dir=task['submit_dir'],
                                      reason=reason)
                if task['pilot'] and task['pilot']['pilot_id']:
                    logger.info("deleting just submitted pilot: %s", task["pilot"]["pilot_id"])
                    await self.rest_client.request('DELETE', f'/pilots/{task["pilot"]["pilot_id"]}')
                if task['pilot'] and 'grid_queue_id' in task['pilot'] and task['pilot']['grid_queue_id']:
                    logger.info("deleting just submitted job: %s", task['pilot']['grid_queue_id'])
                    await self.remove([task['pilot']['grid_queue_id']])
            else:
                if 'grid_queue_id' in task['pilot'] and task['pilot']['grid_queue_id']:
                    grid_queue_ids.append(task['pilot']['grid_queue_id'])

    async def download_input(self, task):
        pass

    async def upload_output(self, task):
        pass

    async def generate_submit_file(self, task, cfg=None, passkey=None,
                                   filelist=None):
        """Generate queueing system submit file for task in dir."""
        args = self.get_submit_args(task,cfg=cfg)
        args.extend(['--offline', '--offline_transfer', 'True', '--gzip-logs'])

        # get requirements and batchopts
        requirements = []
        batch_opts = {}
        input_files = list(filelist)
        output_files = []
        output_remaps = []
        for b in self.queue_cfg['batchopts']:
            if b.lower() == 'requirements':
                requirements.append(self.queue_cfg['batchopts'][b])
            else:
                batch_opts[b] = self.queue_cfg['batchopts'][b]
        if cfg:
            if (cfg['steering'] and 'batchsys' in cfg['steering'] and
                    cfg['steering']['batchsys']):
                for b in cfg['steering']['batchsys']:
                    if self.__class__.__name__.startswith(b.lower()):
                        # these settings apply to this batchsys
                        for bb in cfg['steering']['batchsys'][b]:
                            value = cfg['steering']['batchsys'][b][bb]
                            if bb.lower() == 'requirements':
                                logger.info(f'steering batchsys requirement: {value}')
                                requirements.append(value)
                            else:
                                logger.info(f'steering batchsys other: {bb} = {value}')
                                batch_opts[bb] = value
            if 'task' in cfg['options']:
                t = cfg['options']['task']
                if t in cfg['tasks']:
                    alltasks = [cfg['tasks'][t]]
                else:
                    alltasks = []
                    try:
                        for tt in cfg['tasks']:
                            if t == tt['name']:
                                alltasks.append(tt)
                    except Exception:
                        logger.warning('error finding specified task to run for %r',
                                       task,exc_info=True)
            else:
                alltasks = cfg['tasks']
            logger.info(f'{task["task_id"]} selected tasks: {alltasks}')
            for t in alltasks:
                if 'batchsys' in t and t['batchsys']:
                    logger.info(f'{task["task_id"]} task batchsys: {t["batchsys"]}')
                    for b in t['batchsys']:
                        if self.__class__.__name__.startswith(b.lower()):
                            # these settings apply to this batchsys
                            for bb in t['batchsys'][b]:
                                value = t['batchsys'][b][bb]
                                if bb.lower() == 'requirements':
                                    logger.info(f'{task["task_id"]} task batchsys requirement: {value}')
                                    requirements.append(value)
                                elif bb.lower() == 'transfer_input_files':
                                    input_files.extend(value.split(','))
                                elif bb.lower() == 'transfer_output_files':
                                    output_files.extend(value.split(','))
                                elif bb.lower() == 'transfer_output_remaps':
                                    output_remaps.extend(value.split(','))
                                else:
                                    logger.info(f'{task["task_id"]} task batchsys other: {bb}={value}')
                                    batch_opts[bb] = value

        # write the submit file
        submit_file = os.path.join(task['submit_dir'],'condor.submit')
        with open(submit_file,'w') as f:
            p = partial(print,sep='',file=f)
            p('universe = vanilla')
            p('executable = {}'.format(os.path.join(task['submit_dir'],'loader.sh')))
            p('log = condor.log')
            p('output = condor.out')
            p('error = condor.err')
            p('notification = never')
            p('+IsIceProdJob = True')  # mark as IceProd for monitoring
            p('want_graceful_removal = True')
            if input_files:
                p('transfer_input_files = {}'.format(','.join(input_files)))
            p('skip_filechecks = True')
            p('should_transfer_files = always')
            p('when_to_transfer_output = ON_EXIT_OR_EVICT')
            p('+SpoolOnEvict = False')
            output_files.extend(['iceprod_log.gz', 'iceprod_out.gz', 'iceprod_err.gz'])
            p('transfer_output_files = {}'.format(','.join(output_files)))
            if output_remaps:
                p('transfer_output_remaps = "{}"'.format(';'.join(output_remaps)))

            # put some info about the task in the classads
            p(f'+IceProdDatasetId = "{task["dataset_id"]}"')
            p(f'+IceProdDataset = {task["dataset"]}')
            p(f'+IceProdJobId = "{task["job_id"]}"')
            p(f'+IceProdJobIndex = {task["job"]}')
            p(f'+IceProdTaskId = "{task["task_id"]}"')
            p(f'+IceProdTaskIndex = {task["task_index"]}')
            p(f'+IceProdTaskName = "{task["name"]}"')
            p(f'+IceProdSiteId = "{self.cfg["site_id"]}"')

            # handle resources
            p('+JobIsRunning = (JobStatus =!= 1) && (JobStatus =!= 5)')
            if 'cpu' in task['requirements'] and task['requirements']['cpu']:
                p('request_cpus = {}'.format(task['requirements']['cpu']))
            if 'gpu' in task['requirements'] and task['requirements']['gpu']:
                p('request_gpus = {}'.format(task['requirements']['gpu']))
            else:
                requirements.append('(!isUndefined(Target.GPUs) ? Target.GPUs == 0 : True)')
            if 'memory' in task['requirements'] and task['requirements']['memory']:
                p('request_memory = {}'.format(int(task['requirements']['memory']*1000)))
            if 'disk' in task['requirements'] and task['requirements']['disk']:
                p('request_disk = {}'.format(int(task['requirements']['disk']*1000000)))
            if 'time' in task['requirements'] and task['requirements']['time']:
                p('+OriginalTime = {}'.format(int(task['requirements']['time']*3600)))
                p('+TargetTime = (!isUndefined(Target.PYGLIDEIN_TIME_TO_LIVE) ? Target.PYGLIDEIN_TIME_TO_LIVE : Target.TimeToLive)')
                p('Rank = Rank + (TargetTime - OriginalTime)/86400')
                requirements.append('TargetTime > OriginalTime')
            if 'os' in task['requirements'] and task['requirements']['os']:
                if isinstance(task['requirements']['os'], list):
                    os_type = task['requirements']['os'][0]
                else:
                    os_type = task['requirements']['os']
                if 'singularity' in self.queue_cfg and not self.queue_cfg['singularity']:
                    requirements.append(condor_os_reqs(os_type))
                else:
                    imageoptions = {
                        'RHEL_6_x86_64': 'osgvo-el6:latest',
                        'RHEL_7_x86_64': 'osgvo-el7:latest',
                        'RHEL_8_x86_64': 'osgvo-el8:latest',
                    }
                    if os_type in imageoptions:
                        image = imageoptions[os_type]
                    else:
                        raise Exception(f'bad OS selection: {os_type}')
                    p(f'+SingularityImage="/cvmfs/singularity.opensciencegrid.org/opensciencegrid/{image}"')

            for b in batch_opts:
                p(f'{b}={batch_opts[b]}')
            if requirements:
                p('requirements = ('+')&&('.join(requirements)+')')

            p('arguments = ',' '.join(args))
            p('queue')

    async def submit(self,task):
        """Submit task to queueing system."""
        cmd = ['condor_submit','-terse','condor.submit']
        for tries in range(3):  # make three attempts
            try:
                out = await check_output_clean_env(*cmd, cwd=task['submit_dir'])
                break
            except Exception:
                if tries >= 2:
                    raise
                await asyncio.sleep(1)  # backoff a bit
        grid_queue_id = []
        for line in out.split('\n'):
            # look for range
            parts = [p.strip() for p in line.split('-') if p.strip()]
            if len(parts) != 2:
                continue
            major = parts[0].split('.')[0]
            minor_1 = int(parts[0].split('.')[1])
            minor_2 = int(parts[1].split('.')[1])
            for i in range(minor_1, minor_2+1):
                grid_queue_id.append('{}.{}'.format(major,i))
        task['grid_queue_id'] = ','.join(grid_queue_id)

        return task

    async def get_grid_resources(self, job_id):
        """Get resource information from a running/held task on the queue system"""
        ret = {}
        cmd = ['condor_q', job_id, '-af:,', 'CpusUsage', 'GPUsUsage', 'ResidentSetSize_RAW', 'DiskUsage_RAW', 'LastRemoteWallClockTime']
        out = await check_output_clean_env(*cmd)
        print('get_grid_status():',out)
        cpu, gpu, memory, disk, time = out.strip().split(',')
        if cpu != 'undefined':
            ret['cpu'] = float(cpu)
        if gpu != 'undefined':
            ret['gpu'] = float(gpu)
        if memory != 'undefined':
            ret['memory'] = float(memory)/1024/1024.
        if disk != 'undefined':
            ret['disk'] = float(disk)/1024/1024.
        if time != 'undefined':
            ret['time'] = float(time)/3600.
        return ret

    async def get_grid_status(self):
        """Get all tasks running on the queue system.
           Returns {grid_queue_id:{status,submit_dir}}
        """
        ret = {}
        cmd = ['condor_q', '-constraint', f'Owner == "{getpass.getuser()}" && IceProdSiteId == "{self.cfg["site_id"]}"',
               '-af:j,', 'jobstatus', 'MATCH_EXP_JOBGLIDEIN_ResourceName', 'cmd']
        out = await check_output_clean_env(*cmd)
        print('get_grid_status():',out)
        for line in out.split('\n'):
            if not line.strip():
                continue
            try:
                gid,status,site,cmd = [x.strip() for x in line.split(',') if x.strip()]
            except ValueError:
                logger.warning('bad line: %r', line)
                continue
            if 'loader.sh' not in cmd:
                continue
            if status in ('0', '1'):
                status = 'queued'
            elif status in ('2', '3', '6', '7'):
                status = 'processing'
            elif status == '4':
                status = 'completed'
            elif status in ('5',):
                status = 'error'
            else:
                status = 'unknown'
            ret[gid] = {'status':status,'submit_dir':os.path.dirname(cmd),'site':site}
        return ret

    async def remove(self,tasks):
        """Remove tasks from queueing system."""
        if tasks:
            cmd = ['condor_rm']+list(tasks)
            await check_call_clean_env(*cmd)

    async def get_grid_completions(self):
        """
        Get completions in the last 4 days.

        Returns:
            dict: {grid_queue_id: {status, submit_dir, site} }
        """
        ret = {}
        cmd = ['condor_history', '-constraint', f'Owner == "{getpass.getuser()}" && IceProdSiteId == "{self.cfg["site_id"]}"',
               '-match', '50000', '-af:j,', 'jobstatus', 'exitcode', 'exitbysignal',
               'MATCH_EXP_JOBGLIDEIN_ResourceName', 'cmd']
        out = await check_output_clean_env(*cmd)
        print('get_grid_completions():',out)
        for line in out.split('\n'):
            if not line.strip():
                continue
            try:
                gid,status,exitstatus,exitsignal,site,cmd = [x.strip() for x in line.split(',') if x.strip()]
            except ValueError:
                logger.warning('bad line: %r', line)
                continue
            if 'loader.sh' not in cmd:
                continue
            if status == '4' and exitstatus == '0' and exitsignal == 'false':
                status = 'ok'
            else:
                status = 'error'
            if site == 'undefined':
                site = None
            ret[gid] = {'status':status,'submit_dir':os.path.dirname(cmd),'site':site}
        return ret



"""
Server
======

Run the iceprod server.
"""

import asyncio
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta, UTC
import logging
import os
import sys

from iceprod import __version__ as version_string
from iceprod.core.logger import set_log_level
from iceprod.server.config import IceProdConfig
from iceprod.server.queue import Queue

logger = logging.getLogger('Server')


class Server(object):
    """
    The actual server.

    """
    def __init__(self, config_params=None, outfile=None, errfile=None):
        self.executor = ThreadPoolExecutor(max_workers=10)
        self.cfg = IceProdConfig(override=config_params)
        self.outfile = outfile
        self.errfile = errfile

        self.rotate_logs_task = None
        self.queue = Queue(self.cfg)

        set_log_level(self.cfg['logging']['level'])
        logger.error('IceProd Server - version %s', version_string)

    async def rotate_logs(self):
        current_date = datetime.now(UTC)
        while self.outfile and self.errfile:
            if current_date.day != datetime.now(UTC).day:
                # rotate files
                current_date = datetime.now(UTC)
                if self.outfile:
                    roll_files(sys.stdout, self.outfile)
                if self.errfile:
                    roll_files(sys.stderr, self.errfile)
            await asyncio.sleep(3600)

    async def run(self):
        self.rotate_logs_task = asyncio.create_task(self.rotate_logs())
        try:
            await self.queue.run()
        finally:
            if self.rotate_logs_task:
                self.rotate_logs_task.cancel()
                self.rotate_logs_task = None
            await asyncio.sleep(0)


def roll_files(fd, filename, num_files=5):
    d = datetime.now(UTC)
    ext = (d-timedelta(days=num_files-1)).strftime('%Y-%m-%d')
    newfile = f'{filename}.{ext}'
    if os.path.exists(newfile):  # delete last file
        os.remove(newfile)
    for i in range(num_files-2, 0, -1):
        ext = (d-timedelta(days=i)).strftime('%Y-%m-%d')
        oldfile = f'{filename}.{ext}'
        if os.path.exists(oldfile):
            os.rename(oldfile, newfile)
        newfile = oldfile
    if os.path.exists(filename):
        os.rename(filename, newfile)

    # redirect file descriptor
    newfd = open(filename, fd.mode)
    fd.flush()
    os.dup2(newfd.fileno(), fd.fileno())
    if fd != sys.stdout and fd != sys.stderr:
        fd.close()
    return newfd



import logging
import json
import copy

from jsonschema.exceptions import ValidationError
import tornado.web

from ..base_handler import APIBase
from ..auth import authorization, attr_auth
from iceprod.core.config import Config

logger = logging.getLogger('rest.config')


def setup(handler_cfg):
    """
    Setup method for Config REST API.

    Args:
        handler_cfg (dict): args to pass to the route

    Returns:
        dict: routes, database, indexes
    """
    return {
        'routes': [
            (r'/config/(?P<dataset_id>\w+)', ConfigHandler, handler_cfg),
        ],
        'database': 'config',
        'indexes': {
            'config': {
                'dataset_id_index': {'keys': 'dataset_id', 'unique': True},
            }
        }
    }


class ConfigHandler(APIBase):
    """
    Handle config requests.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id):
        """
        Get a config.

        Args:
            dataset_id (str): the dataset id of the config

        Returns:
            dict: config
        """
        ret = await self.db.config.find_one(
            {'dataset_id':dataset_id},
            projection={'_id':False, 'dataset_id':False}
        )
        if not ret:
            self.send_error(404, reason="Config not found")
        else:
            self.write(ret)

    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='write')
    async def put(self, dataset_id):
        """
        Set a config.

        Body should contain the config.

        Args:
            dataset_id (str): the dataset id of the config

        Returns:
            dict: empty dict
        """
        data = json.loads(self.request.body)
        if 'dataset_id' not in data:
            data['dataset_id'] = dataset_id
        elif data['dataset_id'] != dataset_id:
            raise tornado.web.HTTPError(400, reason='dataset_id mismatch')
        try:
            c = Config(copy.deepcopy(data))
            c.fill_defaults()
            c.validate()
        except ValidationError as e:
            raise tornado.web.HTTPError(400, reason=str(e).split('\n', 1)[0])
        except Exception:
            logger.warning('unknown config validation error', exc_info=True)
            raise tornado.web.HTTPError(400, reason='unknown validation error')
        await self.db.config.replace_one({'dataset_id':dataset_id}, data, upsert=True)
        self.write({})



"""
Reset tasks that are not active (in a pilot).

Check all the tasks that are processing, and compare with the
tasks in pilots.  Reset the difference.
"""

import argparse
import asyncio
from datetime import datetime, UTC
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.server.util import str2datetime

logger = logging.getLogger('non_active_tasks')


async def run(rest_client, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        debug (bool): debug flag to propagate exceptions
    """
    try:
        datasets = await rest_client.request('GET', '/dataset_summaries/status')
        dataset_ids = []
        if 'processing' in datasets:
            dataset_ids.extend(datasets['processing'])
        if 'truncated' in datasets:
            dataset_ids.extend(datasets['truncated'])
        pilots = await rest_client.request('GET', '/pilots', {'keys': 'pilot_id|tasks'})
        task_ids_in_pilots = set()
        for p in pilots.values():
            if 'tasks' in p and p['tasks']:
                task_ids_in_pilots.update(p['tasks'])
        dataset_tasks = {}
        for dataset_id in dataset_ids:
            dataset_tasks[dataset_id] = await rest_client.request('GET', '/datasets/{}/task_summaries/status'.format(dataset_id))

        async def reset(dataset_id, task_id):
            args = {'status':'reset'}
            await rest_client.request('PUT', f'/datasets/{dataset_id}/tasks/{task_id}/status', args)
            data = {
                'name': 'stdlog',
                'task_id': task_id,
                'dataset_id': dataset_id,
                'data': 'task status = processing, but not found in any pilot',
            }
            await rest_client.request('POST', '/logs', data)
            data.update({'name':'stdout', 'data': ''})
            await rest_client.request('POST', '/logs', data)
            data.update({'name':'stderr', 'data': ''})
            await rest_client.request('POST', '/logs', data)

        async def update_pilot(pilot_id, tasks):
            args = {'tasks': tasks}
            await rest_client.request('PATCH', '/pilots/{}'.format(pilot_id), args)

        async def delete_pilot(pilot_id):
            await rest_client.request('DELETE', '/pilots/{}'.format(pilot_id))

        awaitables = set()
        reset_pilots = set()
        now = datetime.now(UTC)
        for dataset_id in dataset_ids:
            tasks = dataset_tasks[dataset_id]
            if 'processing' in tasks:
                reset_tasks = set(tasks['processing'])-task_ids_in_pilots
                for task_id in reset_tasks:
                    args = {'keys': 'status|status_changed'}
                    task = await rest_client.request('GET', f'/datasets/{dataset_id}/tasks/{task_id}', args)
                    # check status, and that we haven't just changed status
                    if task['status'] == 'processing' and (now-str2datetime(task['status_changed'])).total_seconds() > 600:
                        logger.info('dataset %s reset task %s', dataset_id, task_id)
                        awaitables.add(reset(dataset_id,task_id))

            for k in tasks.keys():
                if k in ('reset', 'waiting', 'failed', 'suspended'):
                    for task_id in task_ids_in_pilots.intersection(tasks[k]):
                        args = {'keys': 'status|status_changed'}
                        task = await rest_client.request('GET', f'/datasets/{dataset_id}/tasks/{task_id}', args)
                        # check status, and that we haven't just changed status
                        if task['status'] in ('reset', 'waiting', 'failed', 'suspended') and (now-str2datetime(task['status_changed'])).total_seconds() > 600:
                            reset_pilots.add(task_id)

        for p in pilots.values():
            if 'tasks' in p:
                updated_tasks = set(p['tasks']) - reset_pilots
                if not updated_tasks:
                    awaitables.add(delete_pilot(p['pilot_id']))
                elif updated_tasks != set(p['tasks']):
                    awaitables.add(delete_pilot(p['pilot_id'], updated_tasks))

        for fut in asyncio.as_completed(awaitables):
            try:
                await fut
            except Exception:
                logger.error('error resetting non-active tasks', exc_info=True)
                if debug:
                    raise
    except Exception:
        logger.error('error resetting non-active tasks', exc_info=True)
        if debug:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, debug=args.debug))


if __name__ == '__main__':
    main()



"""
Queue tasks.

Move task statuses from idle to waiting, for a certain number of
tasks.  Also uses priority for ordering.
"""

import argparse
import asyncio
import logging

from wipac_dev_tools import from_environment, strtobool

from iceprod.client_auth import add_auth_to_argparse, create_rest_client

logger = logging.getLogger('queue_tasks')

default_config = {
    'NTASKS': 250000,
    'NTASKS_PER_CYCLE': 1000,
    'TASKS_GET_FACTOR': 5,
}


async def run(rest_client, config, dataset_id='', gpus=None, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        config (dict): config dict
        dataset_id (str): dataset to queue
        gpus (bool): run on gpu tasks, cpu tasks, or both
        debug (bool): debug flag to propagate exceptions
    """
    try:
        num_tasks_idle = 0
        num_tasks_waiting = 0
        if dataset_id:
            route = f'/datasets/{dataset_id}/task_counts/status'
        else:
            route = '/task_counts/status'
        args = {'status': 'idle|waiting'}
        if gpus is not None:
            args['gpu'] = gpus
        tasks = await rest_client.request('GET', route, args)
        if 'idle' in tasks:
            num_tasks_idle = tasks['idle']
        if 'waiting' in tasks:
            num_tasks_waiting = tasks['waiting']
        tasks_to_queue = min(num_tasks_idle, config['NTASKS'] - num_tasks_waiting, config['NTASKS_PER_CYCLE'])
        logger.warning(f'num tasks idle: {num_tasks_idle}')
        logger.warning(f'num tasks waiting: {num_tasks_waiting}')
        logger.warning(f'tasks to waiting: {tasks_to_queue}')

        if tasks_to_queue > 0:
            if dataset_id:
                route = f'/datasets/{dataset_id}/tasks'
                args = {
                    'status': 'idle',
                    'keys': 'task_id|depends|requirements.gpu',
                }
            else:
                route = '/tasks'
                args = {
                    'status': 'idle',
                    'keys': 'task_id|depends|requirements.gpu',
                    'sort': 'priority=-1',
                    'limit': config['TASKS_GET_FACTOR'] * tasks_to_queue,
                }
            ret = await rest_client.request('GET', route, args)
            idle_tasks = ret.values() if dataset_id else ret['tasks']
            queue_tasks = []
            tasks_queue_pending = 0
            deps_futures = set()

            async def check_deps(task):
                for dep in task.get('depends', []):
                    ret = await rest_client.request('GET', f'/tasks/{dep}')
                    if ret['status'] != 'complete':
                        logger.info('dependency not met for task %s', task['task_id'])
                        await rest_client.request('PATCH', f'/tasks/{task["task_id"]}', {'priority': 0})
                        return None
                return task
            for task in idle_tasks:
                if gpus is not None:
                    task_gpus = task.get('requirements', {}).get('gpu', 0)
                    if (gpus and task_gpus <= 0) or (not gpus and task_gpus > 0):
                        continue
                tasks_queue_pending += 1
                if not task.get('depends', None):
                    logger.info('queueing task %s', task['task_id'])
                    queue_tasks.append(task['task_id'])
                else:
                    fut = asyncio.create_task(check_deps(task))
                    deps_futures.add(fut)
                    if len(deps_futures) >= 20:
                        done, pending = await asyncio.wait(deps_futures, return_when=asyncio.FIRST_COMPLETED)
                        for fut in done:
                            task = await fut
                            if task:
                                logger.info('queueing task %s', task['task_id'])
                                queue_tasks.append(task['task_id'])
                            else:
                                tasks_queue_pending -= 1
                        deps_futures = pending
                while tasks_queue_pending >= tasks_to_queue and deps_futures:
                    done, pending = await asyncio.wait(deps_futures, return_when=asyncio.FIRST_COMPLETED)
                    for fut in done:
                        task = await fut
                        if task:
                            logger.info('queueing task %s', task['task_id'])
                            queue_tasks.append(task['task_id'])
                        else:
                            tasks_queue_pending -= 1
                    deps_futures = pending
                if tasks_queue_pending >= tasks_to_queue:
                    break

            if deps_futures:
                for fut in asyncio.as_completed(deps_futures):
                    task = await fut
                    if task:
                        logger.info('queueing task %s', task['task_id'])
                        queue_tasks.append(task['task_id'])

            logger.warning('queueing %d tasks', len(queue_tasks))
            count = 0
            while queue_tasks:
                task_ids = queue_tasks[:100]
                queue_tasks = queue_tasks[100:]
                args = {'task_ids': task_ids}
                ret = await rest_client.request('POST', '/task_actions/waiting', args)
                count += ret['waiting']
            logger.warning('queued %d tasks', count)

    except Exception:
        logger.error('error queueing tasks', exc_info=True)
        if debug:
            raise


def main():
    config = from_environment(default_config)

    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('--dataset-id', help='dataset id')
    parser.add_argument('--gpus', default=None, type=strtobool, help='whether to select only gpu or non-gpu tasks')
    parser.add_argument('--ntasks', type=int, default=config['NTASKS'],
                        help='number of tasks to keep queued')
    parser.add_argument('--ntasks_per_cycle', type=int, default=config['NTASKS_PER_CYCLE'],
                        help='number of tasks to queue per cycle')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()
    config.update(vars(args))

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, dataset_id=args.dataset_id, gpus=args.gpus, config=config, debug=args.debug))


if __name__ == '__main__':
    main()



"""
The schedule module is a basic cron for IceProd, executing tasks
at various intervals or fixed times.
"""
import logging

import iceprod.server
from iceprod.server import module
from iceprod.server.globus import SiteGlobusProxy

logger = logging.getLogger('modules_schedule')


class schedule(module.module):
    """
    Run the schedule module, which handles periodic tasks.

    Tasks live in individual modules in `iceprod.server.scheduled_tasks.*`.
    """
    def start(self):
        super(schedule,self).start()

        try:
            # make sure the gridftp proxy is set up
            proxy_kwargs = {}
            if 'gridftp_cfgfile' in self.cfg['queue']:
                proxy_kwargs['cfgfile'] = self.cfg['queue']['gridftp_cfgfile']
            proxy = SiteGlobusProxy(**proxy_kwargs)
            proxy.update_proxy()
        except Exception:
            logger.warning('Failed to get gridftp proxy')

        # find scheduled tasks, which will run at built-in delays
        task_names = iceprod.server.listmodules('iceprod.server.scheduled_tasks')
        self.tasks = [iceprod.server.run_module(n,self) for n in task_names]



"""
gridftp interface
"""

from __future__ import absolute_import, division, print_function

import os
import logging
from collections import namedtuple
from datetime import datetime
import tempfile
import shutil
import subprocess

logger = logging.getLogger('gridftp')


def _cmd(cmd, timeout=1200):
    subprocess.run(cmd, timeout=timeout, check=True)


def _cmd_output(cmd, timeout=1200):
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    try:
        output = p.communicate(timeout=timeout)[0].decode('utf-8')
        return (p.returncode, output)
    except subprocess.TimeoutExpired:
        p.kill()
        raise Exception('Request timed out')


def listify(lines,details=False,dotfiles=False):
    """Turn ls output into a list of NamedTuples"""
    out = []
    if details:
        File = namedtuple('File', ['directory','perms','subfiles',
                                   'owner','group','size','date',
                                   'name'])
        months = {'jan':1,'feb':2,'mar':3,'apr':4,'may':5,'jun':6,
                  'jul':7,'aug':8,'sep':9,'oct':10,'nov':11,'dec':12}
        for x in lines.split('\n'):
            if not x.strip():
                continue
            pieces = x.split()
            name = os.path.basename(pieces[-1])
            if name.startswith('.') and not dotfiles:
                continue
            d = x[0] == 'd'
            perms = pieces[0][1:]
            subfiles = 0
            if pieces[1].isdigit():
                subfiles = int(pieces[1])
                pieces = pieces[1:]
            else:
                subfiles = int(pieces[3])
            year = datetime.now().year
            month = months[pieces[4].lower()]
            day = int(pieces[5])
            if ':' in pieces[6]:
                hour,minute = pieces[6].split(':')
                dt = datetime(year,month,day,int(hour),int(minute))
            else:
                year = int(pieces[6])
                dt = datetime(year,month,day)
            out.append(File(d,perms,subfiles,pieces[1],pieces[2],
                            int(pieces[3]),dt,name))
    else:
        for x in lines.split('\n'):
            if not x.strip():
                continue
            f = x.split()[-1]
            if not f.startswith('.') or dotfiles:
                out.append(f)
    return out


class GridFTP(object):
    """GridFTP interface to command line client.

       Example:
           GridFTP.get('gsiftp://data.icecube.wisc.edu/file',
                       filename='/path/to/file')
    """

    _timeout = 3600  # 1 hour default timeout

    @classmethod
    def supported_address(cls,address):
        """Return False for address types that are not supported"""
        if '://' not in address:
            return False
        addr_type = address.split(':')[0]
        if addr_type not in ('gsiftp','ftp'):
            return False
        return True

    @classmethod
    def address_split(cls,address):
        """Split an address into server/path parts"""
        pieces = address.split('://',1)
        if '/' in pieces[1]:
            pieces2 = pieces[1].split('/',1)
            return (pieces[0]+'://'+pieces2[0],'/'+pieces2[1])
        else:
            return (address,'/')

    @classmethod
    def get(cls, address, filename=None, request_timeout=None):
        """
        Do a GridFTP get request.

        Either data is returned directly or filename must be defined.

        Args:
            address (str): url to get from
            filename (str): filename to write data to
            request_timeout (float): timeout in secodns

        Returns:
            str: data, if filename is not defined

        Raises:
            Exception for failure
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        tmpdir = None
        if filename is None:
            tmpdir = tempfile.mkdtemp(dir=os.getcwd())
            dest = 'file:'+os.path.join(tmpdir,'get_tmp_file')
        else:
            dest = 'file:'+filename

        cmd = ['globus-url-copy',address,dest]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        try:
            _cmd(cmd, timeout=timeout)
            if filename is None:
                with open(dest[5:]) as f:
                    return f.read()
        finally:
            if tmpdir:
                shutil.rmtree(tmpdir,ignore_errors=True)

    @classmethod
    def put(cls, address, data=None, filename=None, request_timeout=None):
        """
        Do a GridFTP put request.

        Either data or filename must be defined.

        Args:
            address (str): url to put to
            data (str): the data to put
            filename (str): filename for data to put
            request_timeout (float): timeout in seconds

        Raises:
            Exception for failure
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        tmpdir = None
        if data is not None:
            tmpdir = tempfile.mkdtemp(dir=os.getcwd())
            src = 'file:'+os.path.join(tmpdir,'put_tmp_file')
            with open(src[5:],'w' if isinstance(data,str) else 'wb') as f:
                f.write(data)
        elif filename is not None:
            src = 'file:'+filename
        else:
            raise Exception('Neither data or filename is defined')

        cmd = ['globus-url-copy','-cd',src,address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        try:
            _cmd(cmd, timeout=timeout)
        finally:
            if tmpdir:
                shutil.rmtree(tmpdir,ignore_errors=True)

    @classmethod
    def list(cls, address, request_timeout=None, details=False, dotfiles=False):
        """
        Do a GridFTP list request.


        Args:
            address (str): url to list
            request_timeout (float): timeout in seconds
            details (bool): result is a list of NamedTuples
            dotfiles (bool): result includes '.', '..', and other '.' files

        Returns:
           list: a list of files

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-ls',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        if ret[0]:
            if (not dotfiles) and 'No match for' in ret[1]:
                return []
            logger.warning(f'{ret[1]}')
            raise Exception('Error getting listing')
        return listify(ret[1], details=details, dotfiles=dotfiles)

    @classmethod
    def mkdir(cls, address, request_timeout=None, parents=False):
        """
        Make a directory on the ftp server.

        Args:
            address (str): url to directory
            request_timeout (float): timeout in seconds
            parents (bool): make parent directories as needed

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        if parents:
            # recursively make directory
            try:
                cls.mkdir(os.path.basename(address),
                          request_timeout=request_timeout,parents=True)
            except Exception:
                pass

        cmd = ['uberftp','-retry','5','-mkdir',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        _cmd(cmd, timeout=timeout)

    @classmethod
    def rmdir(cls, address, request_timeout=None):
        """
        Remove a directory on the ftp server.

        This fails if the directory is not empty.  Use :py:func:`rmtree` for
        recursive removal.

        Args:
            address (str): url to directory
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-rmdir',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        if ret[0] and 'No match for' not in ret[1]:
            raise Exception('Error removing dir')

    @classmethod
    def delete(cls, address, request_timeout=None):
        """
        Delete a file on the ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-rm',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        if ret[0] and 'No match for' not in ret[1]:
            raise Exception('Error removing dir')

    @classmethod
    def rmtree(cls, address, request_timeout=None):
        """
        Delete a file or directory on the ftp server.

        This is recursive, like `rm -rf`.

        Args:
            address (str): url to file or directory
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-rm','-r',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        if ret[0] and 'No match for' not in ret[1]:
            raise Exception('Error removing dir')

    @classmethod
    def move(cls, src, dest, request_timeout=None):
        """
        Move a file on the ftp server.

        Args:
            src (str): url to source file
            dest (str): url to destination file
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(src):
            raise Exception('address type not supported for src %s'%str(src))
        if not cls.supported_address(dest):
            raise Exception('address type not supported for dest %s'%str(dest))

        cmd = ['uberftp','-retry','5','-rename',src,cls.address_split(dest)[-1]]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        _cmd(cmd, timeout=timeout)

    @classmethod
    def exists(cls, address, request_timeout=None):
        """
        Check if a file exists on the ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-size',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        return (not ret[0])

    @classmethod
    def chmod(cls, address, mode, request_timeout=None):
        """
        Chmod a file on the ftp server.

        Args:
            address (str): url to file
            mode (str): mode of file
            request_timeout (float): timeout in seconds

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-chmod',mode,address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        _cmd(cmd, timeout=timeout)

    @classmethod
    def size(cls, address, request_timeout=None):
        """
        Get the size of a file on the ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Returns:
            int: size of file in bytes

        Raises:
            Exception on error
        """
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))

        cmd = ['uberftp','-retry','5','-size',address]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        ret = _cmd_output(cmd, timeout=timeout)
        if ret[0]:
            raise Exception('failed to get size')
        return int(ret[1])

    @classmethod
    def _chksum(cls, type, address, request_timeout=None):
        """Chksum is faked by redownloading the file and checksumming that"""
        from iceprod.core.functions import cksm
        if not cls.supported_address(address):
            raise Exception('address type not supported for address %s'%str(address))
        if type.endswith('sum'):
            type = type[:-3]

        tmpdir = tempfile.mkdtemp(dir=os.getcwd())
        dest = 'file:'+os.path.join(tmpdir,'dest')

        cmd = ['globus-url-copy',address,dest]

        if request_timeout is None:
            timeout = cls._timeout
        else:
            timeout = request_timeout

        try:
            _cmd(cmd, timeout=timeout)
            if not os.path.exists(dest[5:]):
                raise Exception('failed to redownload')
            return cksm(dest[5:],type)
        finally:
            shutil.rmtree(tmpdir,ignore_errors=True)

    # Some helper functions for different checksum types #

    @classmethod
    def md5sum(cls,address,request_timeout=None):
        """
        Get the md5sum of a file on an ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Returns:
            str: the md5sum

        Raises:
            Exception on error
        """
        return cls._chksum('md5sum',address,request_timeout=request_timeout)

    @classmethod
    def sha1sum(cls,address,request_timeout=None):
        """
        Get the sha1sum of a file on an ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Returns:
            str: the sha1sum

        Raises:
            Exception on error
        """
        return cls._chksum('sha1sum',address,request_timeout=request_timeout)

    @classmethod
    def sha256sum(cls, address, request_timeout=None):
        """
        Get the sha256sum of a file on an ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Returns:
            str: the sha256sum

        Raises:
            Exception on error
        """
        return cls._chksum('sha256sum',address,request_timeout=request_timeout)

    @classmethod
    def sha512sum(cls, address, request_timeout=None):
        """
        Get the sha512sum of a file on an ftp server.

        Args:
            address (str): url to file
            request_timeout (float): timeout in seconds

        Returns:
            str: the sha512sum

        Raises:
            Exception on error
        """
        return cls._chksum('sha512sum',address,request_timeout=request_timeout)



"""
The core execution functions for running on a node.

The fundamental design of the core is to run a task composed of trays and
modules.  The general heirarchy looks like::

    task
    |
    |- tray1
       |
       |- module1
       |
       |- module2
    |
    |- tray2
       |
       |- module3
       |
       |- module4

Parameters can be defined at every level, and each level is treated as a
scope (such that inner scopes inherit from outer scopes).  This is
accomplished via an internal evironment for each scope.

Data movement should be defined at the task level.
"""

from contextlib import contextmanager
import copy
from dataclasses import dataclass
from enum import StrEnum
import logging
import os
from pathlib import Path
from typing import Any, Iterator, Optional


from iceprod.core import config
from iceprod.core.defaults import add_default_options
from iceprod.core import util
from iceprod.core import functions
import iceprod.core.parser
from iceprod.core.jsonUtil import json_encode,json_decode


class ConfigError(Exception):
    pass


class ConfigParser:
    """
    Parse things using a config and the tray/task/module environment.

    Note: dataset config must be valid!

    Args:
        dataset: a dataset object with config
        logger: a logger object, for localized logging
    """
    def __init__(self, dataset: config.Dataset, logger: Optional[logging.Logger] = None):
        dataset.validate()
        self.config = dataset.config
        self.logger = logger if logger else logging.getLogger()
        self.parser = iceprod.core.parser.ExpParser()

    def parseValue(self, value: Any, env: dict = {}) -> Any:
        """
        Parse a value from the available env and global config.

        If the value is a string:
        1. Use the :class:`Meta Parser <iceprod.core.parser>` to parse the string.
        2. Expand any env variables in the result.

        If the value is not a string, pass through the value.

        Args:
            value: the value to parse
            env: tray/task/module env

        Returns:
            the parsed value
        """
        if isinstance(value, str):
            self.logger.debug('parse before:%r| env=%r| options=%r', value, env, self.config.get('options'))
            while value != (ret := self.parser.parse(value, self.config, env)):
                value = ret
            if isinstance(value, str):
                value = os.path.expandvars(value)
            self.logger.debug('parse after:%r', value)
        return value

    def parseObject(self, obj: Any, env: dict) -> Any:
        """
        Recursively parse a dict or list.

        Do not modify original object.

        Args:
            obj: object to parse
            env: tray/task/module env

        Returns:
            the parsed object
        """
        if isinstance(obj, str):
            return self.parseValue(obj, env)
        elif isinstance(obj, (list, tuple)):
            return [self.parseObject(v, env) for v in obj]
        elif isinstance(obj, dict):
            ret = copy.copy(obj)  # use copy.copy in case it's a subclass of dict
            for k in obj:
                ret[k] = self.parseObject(obj[k], env)
            return ret
        else:
            return obj


Env = dict[str, dict[str, Any]]


@contextmanager
def scope_env(cfg: ConfigParser, obj: dict, upperenv: Optional[Env] = None, logger: Optional[logging.Logger] = None) -> Iterator[Env]:
    """
    A context manager for parsing scoped config, such as parameters.

    The returned environment is a dictionary composed of several objects:

    * parameters
        Parameters are defined directly as an object, or as a string pointing
        to another object.  They can use the IceProd meta-language to be
        defined in relation to other parameters specified in inherited
        scopes, or as eval or sprinf functions.

    * input_files
        A set of Data objects (urls and local paths), for files to download before
        the task starts.

    * output_files
        A set of Data objects (urls and local paths), for files to upload after the
        task successfully completes.

    `input_files` and `output_files` are global, while `parameters` is inherited
    at each scope level.

    Args:
        cfg: ConfigParser object
        obj: A partial dataset config section to operate on. The local scope.
        upperenv: previous scope's env output
        logger: a logger object, for localized logging
    """
    env: Env = {
        'parameters': {},
        'input_files': [],
        'output_files': [],
        'upper_input_files': [],
        'upper_output_files': [],
        'environment': {
            'OS_ARCH': '$OS_ARCH',
        }
    }
    if upperenv:
        env['parameters'].update(upperenv['parameters'])
        env['upper_input_files'] = upperenv['input_files'] + upperenv['upper_input_files']
        env['upper_output_files'] = upperenv['output_files'] + upperenv['upper_output_files']

    logger = logger if logger else logging.getLogger()

    try:
        # copy parameters
        if 'parameters' in obj:
            # copy new parameters to env first so local referrals work
            env['parameters'].update(obj['parameters'])
            # parse parameter values and update if necessary
            for p in obj['parameters']:
                newval = cfg.parseValue(obj['parameters'][p], env)
                if newval != obj['parameters'][p]:
                    env['parameters'][p] = newval

        if 'data' in obj:
            # download data
            for data in obj['data']:
                d = cfg.parseObject(data, env)
                if d['movement'] in ('input','both'):
                    ret = downloadData(d, cfg=cfg, logger=logger)
                    if ret:
                        env['input_files'].append(ret)
                if d['movement'] in ('output','both'):
                    ret = uploadData(d, cfg=cfg, logger=logger)
                    if ret:
                        env['output_files'].append(ret)

        # add input and output to parseable options
        cfg.config['options']['input'] = ' '.join(d.local for d in (env['input_files'] + env['upper_input_files']))
        cfg.config['options']['output'] = ' '.join(d.local for d in (env['output_files'] + env['upper_output_files']))
        logging.info('input: %r', cfg.config['options']['input'])
        logging.info('output: %r', cfg.config['options']['output'])

    except util.NoncriticalError:
        logger.warning('Noncritical error when setting up environment', exc_info=True)
    except Exception:
        logger.critical('Serious error when setting up environment', exc_info=True)
        raise

    yield env


class Transfer(StrEnum):
    TRUE = 'true'
    MAYBE = 'maybe'
    FALSE = 'false'


@dataclass(frozen=True, slots=True)
class Data:
    """
    IceProd Data instance

    Args:
        url: url location
        local: local filename
        transfer: whether to transfer file (true | maybe | false)
    """
    url: str
    local: str
    transfer: Transfer

    def __str__(self):
        return f"Data(url='{self.url}', local='{self.local}', transfer='{str(self.transfer)}')"


def storage_location(data: dict, parser: ConfigParser) -> str:
    """
    Get data storage location from the config.

    Args:
        data: data config object
        parser: config parser

    Returns:
        storage location
    """
    config = parser.config
    type_ = data['type'].lower()
    if type_ not in ['permanent', 'job_temp', 'dataset_temp', 'site_temp']:
        raise ConfigError('data movement "type" is unknown')
    if 'options' in config and type_ in config['options']:
        return parser.parseValue(config['options'][type_])
    elif type_ == 'permanent':
        if 'options' in config and 'data_url' in config['options']:
            return parser.parseValue(config['options']['data_url'])
        else:
            raise ConfigError('"data_url" not defined in config["options"]')
    else:
        raise ConfigError(f'{type_} not defined in config["options"]')


def do_transfer(data: dict) -> Transfer:
    """
    Test if we should actually transfer the file.

    Args:
        data: data config object
    """
    ret = Transfer.TRUE
    if isinstance(data['transfer'], bool):
        ret = Transfer.TRUE if data['transfer'] else Transfer.FALSE
    elif isinstance(data['transfer'], str):
        t = data['transfer'].lower()
        if t in ('n', 'no', 'not', 'f', 'false'):
            ret = Transfer.FALSE
        elif t in ('y', 'yes', 't', 'true'):
            ret = Transfer.TRUE
        elif t in ('maybe', 'exists'):
            ret = Transfer.MAYBE
        else:
            raise Exception('unknown transfer type')
    elif isinstance(data['transfer'], (int, float)):
        ret = Transfer.FALSE if data['transfer'] == 0 else Transfer.TRUE
    return ret


def downloadData(data: dict, cfg: ConfigParser, logger=None) -> Optional[Data]:
    """
    Parse download url and local filename.

    Args:
        data: data config object
        cfg: config parser

    Returns:
        either None or a Data object
    """
    if not logger:
        logger = logging
    remote_base = storage_location(data, cfg)
    logger.debug('downloadData(): remote_base: %r', remote_base)
    remote = str(data['remote']) if data['remote'] is not None else ''
    local = str(data['local']) if data['local'] is not None else ''

    if not remote and not local:
        raise ConfigError('need either "remote" or "local" defined for data')
    if not remote:
        url = os.path.join(remote_base, local)
    elif functions.isurl(remote):
        url = remote
    else:
        url = os.path.join(remote_base, remote)

    transfer = do_transfer(data)
    if transfer == Transfer.FALSE:
        logger.info('not transferring file %s', url)
        return

    if not local:
        local = os.path.basename(remote)

    return Data(url, local, transfer)


def uploadData(data: dict, cfg: ConfigParser, logger=None) -> Optional[Data]:
    """
    Parse download url and local filename.

    Args:
        data: data config object
        cfg: config parser

    Returns:
        either None or a Data object
    """
    if not logger:
        logger = logging
    remote_base = storage_location(data, cfg)
    logger.debug('uploadData(): remote_base: %r', remote_base)
    remote = str(data['remote']) if data['remote'] is not None else ''
    local = str(data['local']) if data['local'] is not None else ''

    if not remote and not local:
        raise ConfigError('need either "remote" or "local" defined for data')
    if not remote:
        url = os.path.join(remote_base, local)
    elif not functions.isurl(remote):
        url = os.path.join(remote_base, remote)
    else:
        url = remote

    if not local:
        local = os.path.basename(remote)

    transfer = do_transfer(data)
    if transfer == Transfer.FALSE:
        logger.info('not transferring file %s', local)
        return

    return Data(url, local, transfer)


# Run Functions #

class WriteToScript:
    """
    Write a task to a Bash script, to execute manually.

    Args:
        task: a task object, with dataset config
        workdir: a directory to write the task and any related files
        options: extra dataset config options
        logger: a logger object, for localized logging
    """
    def __init__(self, task: config.Task, workdir: Path, options: Optional[dict] = None, logger: Optional[logging.Logger] = None):
        self.task = task
        self.workdir = workdir
        self.logger = logger if logger else logging.getLogger()

        # default config setup
        self.options = self.task.dataset.config['options']
        self._fill_options()
        if options:
            self.options.update(options)
        self.cfgparser = ConfigParser(self.task.dataset, logger=self.logger)

        # set up script
        self.infiles: set[Data] = set()
        self.outfiles: set[Data] = set()

    def _fill_options(self):
        self.options['dataset_id'] = self.task.dataset.dataset_id
        self.options['dataset'] = self.task.dataset.dataset_num
        self.options['job'] = self.task.job.job_index
        self.options['jobs_submitted'] = self.task.dataset.jobs_submitted
        self.options['task_id'] = self.task.task_id
        self.options['task'] = self.task.name
        self.options['debug'] = self.task.dataset.debug

    def _add_input_files(self, files, f=None):
        if f:
            for data in files:
                if data.transfer is Transfer.FALSE:
                    continue
                if data.url.startswith('gsiftp://'):
                    python_cmd = f'from iceprod.core.gridftp import GridFTP\nGridFTP.get("{data.url}", filename="$PWD/{data.local}")'
                    cmd = [
                        '/cvmfs/icecube.opensciencegrid.org/iceprod/v2.7.1/env-shell.sh',
                        'python', '-', "<<____HERE\n" + python_cmd + '\n____HERE\n',
                    ]
                    print(f'# Input: {data}', file=f)
                    print(' '.join(cmd), file=f)
                    print('', file=f)
                else:
                    self.infiles.add(data)
        else:
            self.infiles.update(files)

    def _add_output_files(self, files, f=None):
        if f:
            for data in files:
                if data.transfer is Transfer.FALSE:
                    continue
                if data.url.startswith('gsiftp://'):
                    python_cmd = f'from iceprod.core.gridftp import GridFTP\nGridFTP.put("{data.url}", filename="$PWD/{data.local}")'
                    cmd_core = [
                        '/cvmfs/icecube.opensciencegrid.org/iceprod/v2.7.1/env-shell.sh',
                        'python', '-', "<<____HERE\n" + python_cmd + '\n____HERE\n',
                    ]
                    if data.transfer is Transfer.MAYBE:
                        cmd = [f'if [ -f {data.local} ]; then\n']
                        cmd.extend(cmd_core)
                        cmd += ['fi']
                    else:
                        cmd = cmd_core
                    print(f'# Output: {data}', file=f)
                    print(' '.join(cmd), file=f)
                    print('', file=f)
                else:
                    self.outfiles.add(data)
        else:
            self.outfiles.update(files)

    async def convert(self, transfer=False):
        """
        Convert to bash script.

        Args:
            transfer: embed the file transfer into the script (default False)
        """
        scriptname = self.workdir / 'task_runner.sh'
        with open(scriptname, 'w') as f:
            print('#!/bin/sh', file=f)
            print('set -e', file=f)
            add_default_options(self.options)
            print('# Options:', file=f)
            for field in self.options:
                print(f'# {field}={self.options[field]}', file=f)
            print('', file=f)
            print('# set some env vars for expansion', file=f)
            print('OS_ARCH=$(/cvmfs/icecube.opensciencegrid.org/py3-v4.3.0/os_arch.sh)', file=f)
            print('', file=f)
            with scope_env(self.cfgparser, self.task.dataset.config['steering'], logger=self.logger) as globalenv:
                task = self.task.get_task_config()
                if self.task.task_files:
                    task['data'].extend(self.task.task_files)
                self._add_input_files(globalenv['input_files'], f=(f if transfer else None))
                self.logger.debug('converting task %s', self.task.name)
                with scope_env(self.cfgparser, task, globalenv, logger=self.logger) as taskenv:
                    self._add_input_files(taskenv['input_files'], f=(f if transfer else None))
                    for i, tray in enumerate(task['trays']):
                        trayname = tray['name'] if tray.get('name', '') else i
                        for iteration in range(tray['iterations']):
                            self.options['iter'] = iteration
                            self.logger.debug('converting tray %r iter %d', trayname, iteration)
                            print(f'# running tray {trayname}, iter {iteration}', file=f)
                            with scope_env(self.cfgparser, tray, taskenv, logger=self.logger) as trayenv:
                                self._add_input_files(trayenv['input_files'], f=(f if transfer else None))
                                for j, module in enumerate(tray['modules']):
                                    modulename = module['name'] if module.get('name', '') else j
                                    self.logger.debug('converting module %r', modulename)
                                    print(f'# running module {modulename}', file=f)
                                    with scope_env(self.cfgparser, module, trayenv, logger=self.logger) as moduleenv:
                                        self._add_input_files(moduleenv['input_files'], f=(f if transfer else None))
                                        await self._write_module(module, moduleenv, file=f)
                                        self._add_output_files(moduleenv['output_files'], f=(f if transfer else None))
                                    print('', file=f)
                                self._add_output_files(trayenv['output_files'], f=(f if transfer else None))
                    self._add_output_files(taskenv['output_files'], f=(f if transfer else None))
                self._add_output_files(globalenv['output_files'], f=(f if transfer else None))

        scriptname.chmod(scriptname.stat().st_mode | 0o700)
        return scriptname

    async def _write_module(self, module, env, file):
        module = module.copy()
        if module['src']:
            module_src = self.cfgparser.parseValue(module['src'], env)
            if functions.isurl(module_src):
                path = os.path.basename(module_src).split('?', 0)[0].split('#', 0)[0]
                env['input_files'].add(Data(
                    url=module_src,
                    local=path,
                    transfer=Transfer.TRUE,
                ))
                module_src = path
            self.logger.info('running module %r with src %s', module['name'], module_src)
        elif module['running_class']:
            module_src = None
            module_class = self.cfgparser.parseValue(module['running_class'], env)
            self.logger.info('running module %r with class %s', module['name'], module_class)
        else:
            self.logger.error('module is missing src')
            raise ConfigError('error running module - need "src"')

        if module['args']:
            module['args'] = self.cfgparser.parseObject(module['args'], env)
        if module['env_shell']:
            module['env_shell'] = self.cfgparser.parseValue(module['env_shell'], env)
        if module['configs']:
            # parse twice to make sure it's parsed, even if it starts as a string
            module['configs'] = self.cfgparser.parseObject(module['configs'], env)
            module['configs'] = self.cfgparser.parseObject(module['configs'], env)

        # set up env_shell
        env_shell = []
        if module['env_shell']:
            env_shell = module['env_shell'].split()
            if functions.isurl(env_shell[0]):
                path = os.path.basename(env_shell[0]).split('?', 0)[0].split('#', 0)[0]
                env['input_files'].add(Data(
                    url=env_shell[0],
                    local=path,
                    transfer=Transfer.TRUE,
                ))
                env_shell[0] = f'./{path}'

        # set up the args
        args = module['args']
        if module_src:
            if args is not None and args != '':
                self.logger.warning('args=%s', args)
                if args and isinstance(args, str) and args[0] in ('{', '['):
                    args = json_decode(args)
                if args and isinstance(args, dict) and set(args) == {'args', 'kwargs'}:
                    args = self.cfgparser.parseObject(args, env)
                elif isinstance(args, str):
                    args = {"args": [self.cfgparser.parseValue(x, env) for x in args.split()], "kwargs": {}}
                elif isinstance(args, list):
                    args = {"args": [self.cfgparser.parseValue(x, env) for x in args], "kwargs": {}}
                elif isinstance(args, dict):
                    args = {"args": [], "kwargs": self.cfgparser.parseObject(args, env)}
                else:
                    args = {"args": [str(args)], "kwargs": {}}

                # convert to cmdline args
                def splitter(a,b):
                    ret = ('-%s' if len(str(a)) <= 1 else '--%s')%str(a)
                    if b is None:
                        return ret
                    else:
                        return ret+'='+str(b)
                args = args['args'] + [splitter(a, args['kwargs'][a]) for a in args['kwargs']]

                # force args to string
                def toStr(a):
                    if isinstance(a,(bytes,str)):
                        return a
                    else:
                        return str(a)
                args = [toStr(a) for a in args]
            else:
                args = []
        else:
            # construct a python file to call the class
            parsed_args = self.cfgparser.parseObject(args, env)
            pymodule, class_ = module_class.rsplit('.', 1)
            args = f"""import json
from {pymodule} import {class_}
args = json.loads('''{json_encode(parsed_args)}''')
obj = {class_}()
for k,v in args.items():
    obj.SetParameter(k, v)
obj.Execute({{}})"""

        # set up the environment
        cmd = []
        if env_shell:
            cmd.extend(env_shell)

        # set up configs
        if module['configs']:
            for filename in module['configs']:
                self.logger.info('creating config %r', filename)
                with open(self.workdir / filename, 'w') as f:
                    f.write(json_encode(module['configs'][filename]))
                data = Data(
                    url=str(self.workdir / filename),
                    local=filename,
                    transfer=Transfer.TRUE,
                )
                env['input_files'].append(data)
                self.infiles.add(data)

        # run the module
        if (not module_src):
            cmd.extend(['python', '-', "<<____HERE\n" + args + '\n____HERE\n'])
        elif module_src[-3:] == '.py':
            # call as python script
            cmd.extend(['python', module_src] + args)
        elif module_src[-3:] == '.sh':
            # call as shell script
            cmd.extend(['/bin/sh', module_src] + args)
        else:
            # call as regular executable
            if module_src[0] != '/':
                module_src = f'./{module_src}'
            cmd.extend([module_src] + args)

        if module['env_clear']:
            # must be on cvmfs-like environ for this to apply
            envstr = 'env -i PYTHONNOUSERSITE=1 '
            for k in ('OPENCL_VENDOR_PATH', 'http_proxy', 'TMP', 'TMPDIR', '_CONDOR_SCRATCH_DIR', 'CUDA_VISIBLE_DEVICES', 'COMPUTE', 'GPU_DEVICE_ORDINAL'):
                envstr += f'{k}=${k} '
            cmd = envstr.split()+cmd

        self.logger.info('cmd=%r',cmd)
        print(' '.join(cmd), file=file)



"""Dataset, job, and task statuses and state machines, and related things."""

from functools import lru_cache, partial

DATASET_STATUS = {
    'processing': ['suspended', 'errors', 'complete'],
    'suspended': ['processing'],
    'errors': ['processing', 'suspended'],
    'complete': [],
}
"""Dataset Statuses

* processing: at least one job is processing.
* suspended: the dataset has been put on hold.
* errors: no jobs are active, and there are jobs with errors.
* complete: all jobs and tasks are complete.
"""

DATASET_STATUS_START = 'processing'
"""Initial dataset status"""

JOB_STATUS = {
    'processing': ['suspended', 'errors', 'complete'],
    'suspended': ['processing'],
    'errors': ['processing'],
    'complete': [],
}
"""Job Statuses

* processing: at least one task is in an active state.
* suspended: a task was put on hold.
* errors: one of the tasks has an error.
* complete: all tasks are complete.
"""

JOB_STATUS_START = 'processing'
"""Initial job status"""

TASK_STATUS = {
    'idle': ['waiting', 'suspended'],
    'waiting': ['queued', 'idle', 'suspended'],
    'queued': ['processing', 'idle', 'waiting', 'suspended', 'failed', 'complete'],
    'processing': ['idle', 'waiting', 'queued', 'suspended', 'failed', 'complete'],
    'suspended': ['idle', 'waiting'],
    'failed': ['idle', 'waiting'],
    'complete': [],
}
"""Task Statuses

* idle: task in IceProd DB, either lower priority or dependencies not met.
* waiting: task in Iceprod DB, ready to run.
* queued: task is queued in HTCondor.
* processing: task is processing in HTCondor.
* suspended: task was manually put on hold.
* failed: task has an error.
* complete: task has completed successfully.
"""

TASK_STATUS_START = 'idle'
"""Initial task status"""


def get_all_prev_statuses(states, status):
    ret = []
    for k,v in states.items():
        if status in v:
            ret.append(k)
    return ret


def status_sort(states, status):
    try:
        return list(states).index(status)
    except ValueError:
        return len(states)


dataset_prev_statuses = lru_cache(partial(get_all_prev_statuses, DATASET_STATUS))
dataset_status_sort = lru_cache(partial(status_sort, DATASET_STATUS))

job_prev_statuses = lru_cache(partial(get_all_prev_statuses, JOB_STATUS))
job_status_sort = lru_cache(partial(status_sort, JOB_STATUS))

task_prev_statuses = lru_cache(partial(get_all_prev_statuses, TASK_STATUS))
task_status_sort = lru_cache(partial(status_sort, TASK_STATUS))



import logging
from functools import wraps
import uuid

import pymongo
from rest_tools.server import catch_error, token_attribute_role_mapping_auth
from tornado.web import HTTPError

from iceprod.roles_groups import ROLES, GROUPS

logger = logging.getLogger('rest-auth')


class AttrAuthMixin:
    async def add_user(self, username):
        """
        Add a user to the auth database.

        Args:
            username (str): username
        """
        user_id = uuid.uuid1().hex
        data = {
            'user_id': user_id,
            'username': username,
            'priority': 0.5,
        }
        await self.auth_db.users.insert_one(data)
        return user_id

    async def set_attr_auth(self, arg, val, read_groups=None, write_groups=None,
                            read_users=None, write_users=None):
        """
        Set the auth for an attribute.  Sets both read and write roles at once.

        Args:
            arg (str): attribute name to set
            val (str): attribute value to set
            read_groups (list): group names of users that should have read access
            write_groups (list): group names of users that should have write access
            read_users (list): user names of users that should have read access
            write_users (list): user names of users that should have write access
        """
        if not read_groups:
            read_groups = []
        if not write_groups:
            write_groups = []
        if not read_users:
            read_users = []
        if not write_users:
            write_users = []
        if any(g for g in read_groups if g not in GROUPS):
            raise Exception(f'Invalid read groups: {read_groups}')
        if any(g for g in write_groups if g not in GROUPS):
            raise Exception(f'Invalid write groups: {write_groups}')

        new_attrs = {'read_groups': read_groups, 'write_groups': write_groups,
                     'read_users': read_users, 'write_users': write_users}

        logger.debug('setting attr auths: arg=%r, val=%r, auths=%r', arg, val, new_attrs)
        ret = await self.auth_db.attr_auths.find_one_and_update(
            {arg: val},
            {'$set': new_attrs},
            upsert=True,
            return_document=pymongo.ReturnDocument.AFTER,
        )
        if ret is None:
            raise RuntimeError('failed to insert auth')

    async def check_attr_auth(self, arg, val, role):
        """
        Based on the request groups or username, check if they are allowed to
        access `arg`:`role`.


        Args:
            arg (str): attribute name to check
            val (str): attribute value
            role (str): the role to check for (read|write)
        """
        try:
            ret = await self.auth_db.attr_auths.find_one({arg: val}, projection={'_id':False})
            if not ret:
                raise HTTPError(403, reason='attr not found')
            elif role+'_groups' not in ret:
                raise HTTPError(403, reason='role not found')
            elif not (set(ret.get(role+'_groups', [])) & set(self.auth_groups) or
                      self.current_user in ret.get(role+'_users', [])):
                logger.debug('arg=%r, val=%r, role=%r, auth_groups=%r, current_user=%r, auths=%r', arg, val, role, self.auth_groups, self.current_user, ret.get(role+'_users', []))
                raise HTTPError(403, reason='authorization failed')
        except (TypeError, ValueError, KeyError):
            logger.debug('arg=%r, val=%r, role=%r, auths=%r', arg, val, role, ret, exc_info=True)
            raise HTTPError(403, reason='failed auth')

    async def manual_attr_auth(self, arg, val, role, token_role_bypass=['admin', 'system']):
        """
        Manually run check_attr_auth and return a boolean.

        Args:
            arg (str): attribute name to check
            val (str): attribute value
            role (str): the role to check for (read|write)
            token_role_bypass (list): token roles that bypass this auth (default: admin,system)

        Returns:
            bool: authorized
        """
        if any(r in self.auth_roles for r in token_role_bypass):
            logger.debug('token role bypass')
            return True
        try:
            await self.check_attr_auth(arg, val, role)
        except Exception:
            logger.debug('unauthorized')
            return False
        return True


#: match token roles and groups
authorization = token_attribute_role_mapping_auth(role_attrs=ROLES, group_attrs=GROUPS)


def attr_auth(**_auth):
    """
    Check attribute auth.  Assume a user is already role-authorized.

    Args:
        arg (str): argument name to look up
        role (str): the role to append to the auth request (read|write)
        token_role_bypass (list): token roles that bypass this auth (default: admin,system)
    """
    def make_wrapper(method):
        @catch_error
        @wraps(method)
        async def wrapper(self, *args, **kwargs):
            arg = _auth['arg']
            role = _auth.get('role', '')
            token_role_bypass = _auth.get('token_role_bypass', ['admin', 'system'])

            if any(r in self.auth_roles for r in token_role_bypass):
                logger.debug('token role bypass')
            else:
                # we need to do an auth check
                val = kwargs.get(arg, None)
                if (not val) or not isinstance(val, str):
                    raise HTTPError(403, reason='authorization failed')
                await self.check_attr_auth(arg, val, role)

            return await method(self, *args, **kwargs)
        return wrapper
    return make_wrapper



import argparse
import asyncio
import logging
import time

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.core.parser import ExpParser
from iceprod.core.resources import Resources
from iceprod.server.priority import Priority
from iceprod.server.states import TASK_STATUS

logger = logging.getLogger('materialize')

DATASET_CYCLE_TIMEOUT = 120


class Materialize:
    def __init__(self, rest_client):
        self.rest_client = rest_client
        self.config_cache = {}
        self.prio = None

    async def run_once(self, only_dataset=None, set_status=None, num=10000, dryrun=False):
        """
        Actual materialization work.

        Args:
            only_dataset (str): dataset_id if we should only buffer a single dataset
            set_status (str): status of new tasks
            num (int): max number of jobs to buffer
            dryrun (bool): if true, do not modify DB, just log changes
        """
        if set_status and set_status not in TASK_STATUS:
            raise Exception('set_status is not a valid task status')
        self.config_cache = {}  # clear config cache
        self.prio = Priority(self.rest_client)  # clear priority cache

        ret = True

        if only_dataset:
            datasets = [only_dataset]
        else:
            ret = await self.rest_client.request('GET', '/dataset_summaries/status')
            datasets = ret.get('processing', [])

        for dataset_id in datasets:
            try:
                start_time = time.time()
                dataset = await self.rest_client.request('GET', '/datasets/{}'.format(dataset_id))
                if dataset.get('truncated', False) and not only_dataset:
                    logger.info('ignoring truncated dataset %s', dataset_id)
                    continue
                job_counts = await self.rest_client.request('GET', '/datasets/{}/job_counts/status'.format(dataset_id))
                tasks = await self.rest_client.request('GET', '/datasets/{}/task_counts/status'.format(dataset_id))
                if 'waiting' not in tasks or job_counts.get('processing', 0) < num or only_dataset:
                    # buffer for this dataset
                    logger.warning('checking dataset %s', dataset_id)
                    jobs = await self.rest_client.request('GET', '/datasets/{}/jobs'.format(dataset_id), {'keys': 'job_id|job_index'})
                    job_index_id = {j['job_index']: j['job_id'] for j in jobs.values()}

                    # check that last job was buffered correctly
                    job_index = max(jobs[i]['job_index'] for i in jobs)+1 if jobs else 0
                    num_tasks = sum(tasks.values())
                    logger.info('job_index: %d', job_index)
                    logger.info('num_tasks: %d', num_tasks)
                    logger.info('tasks_per_job: %d', dataset['tasks_per_job'])
                    for job_tmp_index in range(job_index-1, -1, -1):
                        if job_index * dataset['tasks_per_job'] <= num_tasks:
                            break
                        logger.info('a job must have failed to buffer, so check in reverse order. job_index=%d, num_tasks=%d', job_tmp_index, num_tasks)
                        job_tasks = await self.rest_client.request('GET', f'/datasets/{dataset_id}/tasks',
                                                                   {'job_index': job_tmp_index, 'keys': 'task_id|job_id|task_index'})
                        if len(job_tasks) != dataset['tasks_per_job']:
                            logger.info('fixing buffer of job %d for dataset %s', job_tmp_index, dataset_id)
                            job_id = job_index_id[job_tmp_index]
                            logger.info('  fixing job_id %s, num existing tasks: %d', job_id, len(job_tasks))
                            tasks_buffered = await self.buffer_job(dataset, job_tmp_index, job_id=job_id,
                                                                   tasks=list(job_tasks.values()),
                                                                   set_status=set_status, dryrun=dryrun)
                            num_tasks += tasks_buffered
                            logger.info('buffered %d tasks. num_tasks increased to: %d', tasks_buffered, num_tasks)

                    # now try buffering new tasks
                    jobs_to_buffer = min(num, dataset['jobs_submitted'] - job_index)
                    if jobs_to_buffer > 0:
                        logger.info('buffering %d jobs for dataset %s', jobs_to_buffer, dataset_id)
                        for i in range(jobs_to_buffer):
                            await self.buffer_job(dataset, job_index, set_status=set_status, dryrun=dryrun)
                            job_index += 1

                            if only_dataset is None and time.time() - start_time > DATASET_CYCLE_TIMEOUT:
                                logger.warning('dataset cycle timeout for dataset %s', dataset_id)
                                ret = False
                                break
            except Exception:
                logger.error('error buffering dataset %s', dataset_id, exc_info=True)
                if only_dataset:
                    raise

        return ret

    async def buffer_job(self, dataset, job_index, job_id=None, tasks=None, set_status=None, dryrun=False):
        """
        Buffer a single job for a dataset

        Args:
            dataset (dict): dataset info
            job_index (int): job index
            job_id (str): job id (if filling in remaining)
            tasks (list): existing tasks (if filling in remaining)
            set_status (str): status of new tasks
            dryrun (bool): set to True if this is a dry run

        Returns:
            int: number of tasks buffered
        """
        dataset_id = dataset['dataset_id']
        logger.info('buffering dataset %s job %d', dataset_id, job_index)

        config = await self.get_config(dataset_id)
        parser = ExpParser()
        task_names = [task['name'] if task['name'] else str(i) for i,task in enumerate(config['tasks'])]
        if len(task_names) != dataset['tasks_per_job']:
            raise Exception('config num tasks does not match dataset tasks_per_job')

        args = {'dataset_id': dataset_id, 'job_index': job_index}
        if dryrun:
            job_id = {'result': 'DRYRUN'}
            task_ids = []
            task_iter = list(enumerate(task_names))
        elif job_id:
            task_ids = [task['task_id'] for task in tasks] if tasks else []
            task_indexes = {task['task_index'] for task in tasks} if tasks else {}
            task_iter = [(i,name) for i,name in enumerate(task_names) if i not in task_indexes]
            if not task_iter:
                raise Exception('no task names to create')
        else:
            ret = await self.rest_client.request('POST', '/jobs', args)
            job_id = ret['result']
            task_ids = []
            task_iter = list(enumerate(task_names))

        # buffer tasks
        for task_index,name in task_iter:
            logger.info('  buffering task_index %d, name %s', task_index, name)
            depends = await self.get_depends(config, job_index,
                                             task_index, task_ids)
            config['options']['job'] = job_index
            config['options']['task'] = task_index
            config['options']['dataset'] = dataset['dataset']
            config['options']['jobs_submitted'] = dataset['jobs_submitted']
            config['options']['tasks_submitted'] = dataset['tasks_submitted']
            config['options']['debug'] = dataset['debug']
            args = {
                'dataset_id': dataset_id,
                'job_id': job_id,
                'task_index': task_index,
                'job_index': job_index,
                'name': name,
                'depends': depends,
                'requirements': self.get_reqs(config, task_index, parser),
            }
            if set_status:
                args['status'] = set_status
            if dryrun:
                logger.info(f'DRYRUN: POST /tasks {args}')
                task_ids.append(job_index*1000000+task_index)
            else:
                ret = await self.rest_client.request('POST', '/tasks', args)
                task_id = ret['result']
                task_ids.append(task_id)
                p = await self.prio.get_task_prio(dataset_id, task_id)
                await self.rest_client.request('PATCH', f'/tasks/{task_id}', {'priority': p})

        return len(task_iter)

    async def get_config(self, dataset_id):
        """Get dataset config"""
        if dataset_id in self.config_cache:
            return self.config_cache[dataset_id]

        config = await self.rest_client.request('GET', '/config/{}'.format(dataset_id))
        if 'options' not in config:
            config['options'] = {}
        self.config_cache[dataset_id] = config
        return config

    def get_reqs(self, config, task_index, parser):
        """
        Get requirements for a task.

        Args:
            config (:py:class:`iceprod.core.dataclasses.Job`): dataset config
            task_index (int): task index
            parser (:py:class:`iceprod.core.parser.ExpParser`): parser

        Returns:
            dict: task requirements
        """
        task = config['tasks'][task_index]
        req = task['requirements'].copy() if 'requirements' in task else {}
        for k in req:
            if isinstance(req[k], list):
                req[k] = [parser.parse(val,config) for val in req[k]]
            else:
                req[k] = parser.parse(req[k],config)
        for k in Resources.defaults:
            if k == 'gpu' and k in req and isinstance(req[k], (tuple,list)):
                req[k] = len(req[k])
        return req

    async def get_depends(self, config, job_index, task_index, task_ids):
        """
        Get dependency task_ids for a task.

        Args:
            config (:py:class:`iceprod.core.dataclasses.Job`): dataset config
            job_index (int): job index
            task_index (int): task index
            task_ids (list): list of already buffered task_ids in this job

        Returns:
            list: list of task_id dependencies
        """
        task = config['tasks'][task_index]
        if 'depends' not in task:
            return []
        all_tasks = {task['name'] if task['name'] else str(i):i for i,task in enumerate(config['tasks'])}
        ret = []
        for dep in task['depends']:
            if dep in all_tasks:
                logging.debug('dep %r in all_tasks', dep)
                dep_index = all_tasks[dep]
                if dep_index >= task_index:
                    raise Exception("bad depends: %r. can't depend on ourself or a greater index", dep)
                ret.append(task_ids[dep_index])
            elif isinstance(dep, int) and dep < len(all_tasks):
                logging.debug('dep %r is int', dep)
                dep_index = dep
                if dep_index >= task_index:
                    raise Exception("bad depends: %r. can't depend on ourself or a greater index", dep)
                ret.append(task_ids[dep_index])
            elif isinstance(dep,str) and ':' in dep:
                # try for a task in another dataset
                logging.debug('dep %r in another dataset', dep)
                dataset_id, dep = dep.split(':',1)
                try:
                    tasks = await rest_client.request('GET', '/datasets/{}/tasks?keys=task_id|name|task_index&job_index={}'.format(dataset_id,job_index))
                    for task in tasks.values():
                        if dep == task['name'] or dep == str(task['task_index']):
                            ret.append(task['task_id'])
                            break
                    else:
                        raise Exception()
                except Exception:
                    raise Exception('bad depends: %r', dep)
            else:
                # try for a raw task_id
                logging.debug('dep %r is a raw id', dep)
                try:
                    task = await self.rest_client.request('GET', '/tasks/{}'.format(dep))
                    ret.append(task['task_id'])
                except Exception:
                    raise Exception('bad depends: %r', dep)

        return ret


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Materialize a dataset')
    parser.add_argument('dataset_id')
    add_auth_to_argparse(parser)
    parser.add_argument('--set_status', default=None, help='initial task status')
    parser.add_argument('-n', '--num', default=100, type=int, help='number of jobs to materialize')
    parser.add_argument('--job_index', type=int, help='specific job index to buffer')
    parser.add_argument('--job_id', default=None, help='specific job id to buffer tasks into')
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--dryrun', action='store_true', help='do not modify database, just log changes')
    args = parser.parse_args()
    logging.basicConfig(level=(logging.DEBUG if args.debug else logging.INFO))

    rest_client = create_rest_client(args)
    materialize = Materialize(rest_client)
    if args.job_index is not None:
        logging.warning('manually buffering a job for dataset %s job %d', args.dataset_id, args.job_index)

        async def run():
            dataset = await rest_client.request('GET', '/datasets/{}'.format(args.dataset_id))
            materialize.prio = Priority(rest_client)
            await materialize.buffer_job(dataset, args.job_index, job_id=args.job_id)

        asyncio.run(run())
    else:
        asyncio.run(materialize.run_once(only_dataset=args.dataset_id, set_status=args.set_status, num=args.num, dryrun=args.dryrun))



"""
Buffer jobs and tasks into the queue.

Also known as "late materialization" of jobs and tasks,
this method finds datasets that are not fully materialized and
buffers a few more jobs and tasks into existence so they can be queued.
"""

import argparse
import asyncio
import logging

from iceprod.client_auth import add_auth_to_argparse, create_rest_client

logger = logging.getLogger('buffer_jobs_tasks')


async def run(rest_client, only_dataset=None, num=1000, run_once=False, delay=10, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        only_dataset (str): dataset_id if we should only buffer a single dataset
        num (int): max number of jobs per dataset to buffer
        run_once (bool): flag to only run once and stop
        delay (float): wait N seconds between result checks
        debug (bool): debug flag to propagate exceptions
    """
    try:
        logger.info('starting materialization request')
        if only_dataset:
            ret = await rest_client.request('POST', f'/request/{only_dataset}', {'num': num})
        else:
            ret = await rest_client.request('POST', '/', {'num': num})
        materialization_id = ret['result']
        logger.info(f'waiting for materialization request {materialization_id}')

        while True:
            await asyncio.sleep(delay)
            ret = await rest_client.request('GET', f'/status/{materialization_id}')
            if ret['status'] == 'complete':
                logger.info(f'materialization request {materialization_id} complete')
                break
            elif ret['status'] == 'error':
                logger.warning(f'materialization request {materialization_id} failed')
                if run_once:
                    raise Exception('materialization failed: %r', ret)
                break
    except Exception:
        logger.warning('materialization error', exc_info=True)
        if debug or run_once:
            raise


def main():
    parser = argparse.ArgumentParser(description='run a scheduled task once')
    add_auth_to_argparse(parser)
    parser.add_argument('-d', '--dataset', type=str, default=None, help='dataset id (optional)')
    parser.add_argument('-n', '--num', type=int, default=100, help='number of jobs per dataset to buffer')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    asyncio.run(run(rest_client, only_dataset=args.dataset, num=args.num, run_once=True, debug=args.debug))


if __name__ == '__main__':
    main()



"""
Logfile setup
"""

from __future__ import absolute_import, division, print_function

import os
import time
import logging
import logging.handlers
import gzip

setlevel = {
    'CRITICAL': logging.CRITICAL,  # execution cannot continue
    'FATAL': logging.CRITICAL,
    'ERROR': logging.ERROR,  # something is wrong, but try to continue
    'WARNING': logging.WARNING,  # non-ideal behavior, important event
    'WARN': logging.WARNING,
    'INFO': logging.INFO,  # initial debug information
    'DEBUG': logging.DEBUG  # the things no one wants to see
}

host = os.uname()[1].split(".")[0]


def set_logger(loglevel='INFO', logfile=None, timedrotate=True, logsize=2**28, lognum=6):
    """Add an output to the root logger"""
    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'

    rootLogger = logging.getLogger()

    if logfile:
        if not logfile.startswith('/'):
            if 'I3PROD' in os.environ:
                logfile = os.path.expanduser(os.path.expandvars(
                    os.path.join('$I3PROD', 'var', 'log', logfile)))
            else:
                logfile = os.path.join(os.getcwd(), 'log', host, logfile)
        if not os.path.exists(os.path.dirname(logfile)):
            os.makedirs(os.path.dirname(logfile))
        if logfile.endswith('.gz'):
            z_file = gzip.open(logfile, mode='wt', encoding='utf-8')
            fileHandler = logging.StreamHandler(z_file)
        elif timedrotate:
            fileHandler = logging.handlers.TimedRotatingFileHandler(logfile, when='midnight', backupCount=lognum, utc=True)
        else:
            fileHandler = logging.handlers.RotatingFileHandler(logfile, mode='a', maxBytes=logsize, backupCount=lognum)
        formatter = logging.Formatter(logformat)
        formatter.converter = time.gmtime
        fileHandler.setFormatter(formatter)
        rootLogger.addHandler(fileHandler)
        for handler in rootLogger.handlers:
            if handler != fileHandler:
                rootLogger.removeHandler(handler)
        rootLogger.setLevel(setlevel[loglevel.upper()])
        rootLogger.info('fileHandler used')
    else:
        logging.basicConfig(format=logformat, level=setlevel[loglevel.upper()])
        rootLogger.info('basicConfig used')

    rootLogger.info('loglevel %s, logfile %s, logsize %d, lognum %d',
                    loglevel, logfile, logsize, lognum)


def set_log_level(loglevel='INFO'):
    rootLogger = logging.getLogger()
    rootLogger.setLevel(setlevel[loglevel.upper()])


def new_file(filename):
    """Write logging to a new file"""
    log = logging.getLogger()
    handlers = False
    for handler in log.handlers:
        if isinstance(handler, logging.handlers.RotatingFileHandler):
            new_handler = logging.handlers.RotatingFileHandler(filename, 'a',
                                                               handler.maxBytes,
                                                               handler.backupCount)
            new_handler.setFormatter(handler.formatter)
            log.addHandler(new_handler)
            log.removeHandler(handler)
            handlers = True
    if not handlers:
        set_logger(logfile=filename)
    logging.info('loggers=%r', log.handlers)


def remove_stdout():
    """Remove the stdout log output from the root logger"""
    log = logging.getLogger()
    logging.info('remove_stdout(): loggers=%s', log.handlers)
    for handler in log.handlers:
        if isinstance(handler,logging.StreamHandler):
            log.removeHandler(handler)
    logging.info('loggers=%s', log.handlers)


def remove_handlers():
    """Remove all handlers from the root logger"""
    log = logging.getLogger()
    logging.info('remove_handlers(): loggers=%s', log.handlers)
    for handler in log.handlers:
        log.removeHandler(handler)
    logging.info('loggers=%s', log.handlers)


def rotate():
    """Rotate the file in the root logger"""
    log = logging.getLogger()
    logging.info('rotate() loggers=%s', log.handlers)
    for handler in log.handlers:
        if isinstance(handler,logging.handlers.RotatingFileHandler):
            handler.doRollover()



"""
Materialization service for dataset late materialization.
"""
import asyncio
from datetime import datetime
import json
import logging
import time
import uuid

import pymongo
import pymongo.errors
import motor.motor_asyncio
import requests.exceptions
from rest_tools.client import RestClient, ClientCredentialsAuth
from rest_tools.server import RestServer
from tornado.web import HTTPError
from tornado.web import RequestHandler as TornadoRequestHandler
from wipac_dev_tools import from_environment

from iceprod.rest.auth import authorization, attr_auth
from iceprod.rest.base_handler import IceProdRestConfig, APIBase
from iceprod.server.module import FakeStatsClient, StatsClientIgnoreErrors
from iceprod.server.util import nowstr, datetime2str
from .service import MaterializationService

logger = logging.getLogger('server')


class BaseHandler(APIBase):
    def initialize(self, materialization_service=None, rest_client=None, **kwargs):
        super().initialize(**kwargs)
        self.materialization_service = materialization_service
        self.rest_client = rest_client

    async def new_request(self, args):
        # validate first
        fields = {
            'dataset_id': str,
            'set_status': str,
            'num': int,
        }
        if set(args)-set(fields):  # don't let random args through
            raise HTTPError(400, reason='invalid params')
        for k in fields:
            if k in args and not isinstance(args[k], fields[k]):
                r = 'key "{}" should be of type {}'.format(k, fields[k].__name__)
                raise HTTPError(400, reason=r)

        # set some fields
        now = nowstr()
        data = {
            'materialization_id': uuid.uuid1().hex,
            'status': 'waiting',
            'create_timestamp': now,
            'modify_timestamp': now,
            'creator': self.auth_data.get('username', None),
            'role': self.auth_data.get('role', None),
        }
        for k in fields:
            if k in args:
                data[k] = args[k]

        # insert
        await self.db.materialization.insert_one(data)
        return data

    async def check_attr_auth(self, arg, val, role):
        """
        Based on the request groups or username, check if they are allowed to
        access `arg`:`role`.

        Runs a remote query to the IceProd API.

        Args:
            arg (str): attribute name to check
            val (str): attribute value
            role (str): the role to check for (read|write)
        """
        args = {
            'name': arg,
            'value': val,
            'role': role,
            'username': self.current_user,
            'groups': self.auth_groups,
        }
        try:
            await self.rest_client.request('POST', '/auths', args)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 403:
                raise HTTPError(403, 'auth failed')
            else:
                raise HTTPError(500, 'auth could not be completed')


class StatusHandler(BaseHandler):
    """
    Handle materialization status requests.
    """
    @authorization(roles=['admin', 'system'])
    async def get(self, materialization_id):
        """
        Get materialization status.

        If materialization_id is invalid, returns http code 404.

        Args:
            materialization_id (str): materialization request id

        Returns:
            dict: materialization metadata
        """
        ret = await self.db.materialization.find_one({'materialization_id':materialization_id},
                                                     projection={'_id':False})
        if not ret:
            self.send_error(404, reason="Materialization request not found")
        else:
            self.write(ret)


class RequestHandler(BaseHandler):
    """
    Handle basic materialization requests.
    """
    @authorization(roles=['admin', 'system'])
    async def post(self):
        """
        Create basic materialization request.

        On success, returns http code 201.

        Params:
            num: number of jobs to buffer per dataset

        Returns:
            dict: {result: materialization_id}
        """
        args = json.loads(self.request.body) if self.request.body else {}

        data = await self.new_request(args)

        # return success
        self.set_status(201)
        self.write({'result': data['materialization_id']})


class RequestDatasetHandler(BaseHandler):
    """
    Handle dataset materialization requests.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='write')
    async def post(self, dataset_id):
        """
        Create dataset materialization request.

        On success, returns http code 201.

        Params:
            num: number of jobs to buffer per dataset

        Args:
            dataset_id (str): dataset_id to materialize

        Returns:
            dict: {result: materialization_id}
        """
        args = json.loads(self.request.body) if self.request.body else {}
        args['dataset_id'] = dataset_id

        data = await self.new_request(args)

        # return success
        self.set_status(201)
        self.write({'result': data['materialization_id']})


class DatasetStatusHandler(BaseHandler):
    """
    Handle dataset materialization request statuses.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id):
        """
        Get latest materialization status for a dataset, if any.

        If dataset_id is invalid, returns http code 404.

        Args:
            dataset_id (str): dataset id to check

        Returns:
            dict: materialization metadata
        """
        ret = await self.db.materialization.find_one(
            {'dataset_id': dataset_id},
            projection={'_id':False},
            sort=[('modify_timestamp', pymongo.DESCENDING)],
        )
        if not ret:
            self.send_error(404, reason="Materialization request not found")
        else:
            self.write(ret)


class HealthHandler(BaseHandler):
    """
    Handle health requests.
    """
    async def get(self):
        """
        Get health status.

        Returns based on exit code, 200 = ok, 400 = failure
        """
        now = time.time()
        status = {
            'now': nowstr(),
            'start_time': datetime2str(datetime.utcfromtimestamp(self.materialization_service.start_time)),
            'last_run_time': "",
            'last_success_time': "",
            'last_cleanup_time': "",
            'num_requests': -1,
        }
        try:
            if self.materialization_service.last_run_time is None and self.materialization_service.start_time + 3600 < now:
                self.send_error(500, reason='materialization was never run')
                return
            if self.materialization_service.last_run_time is not None:
                if self.materialization_service.last_run_time + 3600 < now:
                    self.send_error(500, reason='materialization has stopped running')
                    return
                status['last_run_time'] = datetime2str(datetime.utcfromtimestamp(self.materialization_service.last_run_time))
            if self.materialization_service.last_success_time is None and self.materialization_service.start_time + 86400 < now:
                self.send_error(500, reason='materialization was never successful')
                return
            if self.materialization_service.last_success_time is not None:
                if self.materialization_service.last_success_time + 86400 < now:
                    self.send_error(500, reason='materialization has stopped being successful')
                    return
                status['last_success_time'] = datetime2str(datetime.utcfromtimestamp(self.materialization_service.last_success_time))
            if self.materialization_service.last_cleanup_time is not None:
                status['last_cleanup_time'] = datetime2str(datetime.utcfromtimestamp(self.materialization_service.last_cleanup_time))
        except Exception:
            logger.info('error from materialization service', exc_info=True)
            self.send_error(500, reason='error from materialization service')
            return

        try:
            ret = await self.db.materialization.count_documents({'status':{'$in':['waiting','processing']}}, maxTimeMS=1000)
        except Exception:
            logger.info('bad db request', exc_info=True)
            self.send_error(500, reason='bad db request')
            return
        if ret is None:
            self.send_error(500, reason='bad db result')
        else:
            status['num_requests'] = ret
            self.write(status)


class Error(TornadoRequestHandler):
    def prepare(self):
        raise HTTPError(404, 'invalid api route')


class Server:
    def __init__(self):
        default_config = {
            'HOST': 'localhost',
            'PORT': 8080,
            'DEBUG': False,
            'OPENID_URL': '',
            'OPENID_AUDIENCE': '',
            'ICEPROD_API_ADDRESS': 'https://iceprod2-api.icecube.wisc.edu',
            'ICEPROD_API_CLIENT_ID': '',
            'ICEPROD_API_CLIENT_SECRET': '',
            'DB_URL': 'mongodb://localhost/datasets',
            'DB_TIMEOUT': 60,
            'DB_WRITE_CONCERN': 1,
            'STATSD_ADDRESS': '',
            'STATSD_PREFIX': 'rest_api',
            'CI_TESTING': '',
        }
        config = from_environment(default_config)

        rest_config = {
            'debug': config['DEBUG'],
        }
        if config['OPENID_URL']:
            logging.info(f'enabling auth via {config["OPENID_URL"]} for aud "{config["OPENID_AUDIENCE"]}"')
            rest_config.update({
                'auth': {
                    'openid_url': config['OPENID_URL'],
                    'audience': config['OPENID_AUDIENCE'],
                }
            })
        elif config['CI_TESTING']:
            rest_config.update({
                'auth': {
                    'secret': 'secret',
                }
            })
        else:
            raise RuntimeError('OPENID_URL not specified, and CI_TESTING not enabled!')

        statsd = FakeStatsClient()
        if config['STATSD_ADDRESS']:
            try:
                addr = config['STATSD_ADDRESS']
                port = 8125
                if ':' in addr:
                    addr,port = addr.split(':')
                    port = int(port)
                statsd = StatsClientIgnoreErrors(addr, port=port, prefix=config['STATSD_PREFIX'])
            except Exception:
                logger.warning('failed to connect to statsd: %r', config['STATSD_ADDRESS'], exc_info=True)

        logging_url = config["DB_URL"].split('@')[-1] if '@' in config["DB_URL"] else config["DB_URL"]
        logging.info(f'DB: {logging_url}')
        db_url, db_name = config['DB_URL'].rsplit('/', 1)
        db = motor.motor_asyncio.AsyncIOMotorClient(
            db_url,
            timeoutMS=config['DB_TIMEOUT']*1000,
            w=config['DB_WRITE_CONCERN'],
        )
        logging.info(f'DB name: {db_name}')
        self.db = db[db_name]
        self.indexes = {
            'materialization': {
                'materialization_id_index': {'keys': 'materialization_id', 'unique': True},
                'dataset_id_index': {'keys': 'dataset_id', 'unique': False},
                'status_timestamp_index': {'keys': [('status', pymongo.ASCENDING), ('timestamp', pymongo.ASCENDING)], 'unique': False},
            }
        }

        if config['ICEPROD_API_CLIENT_ID'] and config['ICEPROD_API_CLIENT_SECRET']:
            logging.info(f'enabling auth via {config["OPENID_URL"]} for aud "{config["OPENID_AUDIENCE"]}"')
            rest_client = ClientCredentialsAuth(
                address=config['ICEPROD_API_ADDRESS'],
                token_url=config['OPENID_URL'],
                client_id=config['ICEPROD_API_CLIENT_ID'],
                client_secret=config['ICEPROD_API_CLIENT_SECRET'],
            )
        elif config['CI_TESTING']:
            rest_client = RestClient(config['ICEPROD_API_ADDRESS'], timeout=1, retries=0)
        else:
            raise RuntimeError('ICEPROD_API_CLIENT_ID or ICEPROD_API_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

        self.materialization_service = MaterializationService(self.db, rest_client)
        self.materialization_service_task = None

        kwargs = IceProdRestConfig(rest_config, statsd=statsd, database=self.db)
        kwargs['materialization_service'] = self.materialization_service
        kwargs['rest_client'] = rest_client

        server = RestServer(debug=config['DEBUG'])

        server.add_route(r'/status/(?P<materialization_id>\w+)', StatusHandler, kwargs)
        server.add_route('/', RequestHandler, kwargs)
        server.add_route(r'/request/(?P<dataset_id>\w+)', RequestDatasetHandler, kwargs)
        server.add_route(r'/request/(?P<dataset_id>\w+)/status', DatasetStatusHandler, kwargs)
        server.add_route('/healthz', HealthHandler, kwargs)
        server.add_route(r'/(.*)', Error)

        server.startup(address=config['HOST'], port=config['PORT'])

        self.server = server

    async def start(self):
        for collection in self.indexes:
            existing = await self.db[collection].index_information()
            for name in self.indexes[collection]:
                if name not in existing:
                    logging.info('DB: creating index %s:%s', collection, name)
                    kwargs = self.indexes[collection][name]
                    await self.db[collection].create_index(name=name, **kwargs)

        if not self.materialization_service_task:
            self.materialization_service_task = asyncio.create_task(self.materialization_service.run())

    async def stop(self):
        await self.server.stop()
        if self.materialization_service_task:
            self.materialization_service_task.cancel()
            try:
                await self.materialization_service_task
            except asyncio.CancelledError:
                pass
            finally:
                self.materialization_service_task = None
